##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from datetime import timezone
import logging
import os
import pickle

import appdirs
import cachetools
import pkg_resources

from photini.imagelist import DRAG_MIMETYPE
from photini.pyqt import *
from photini.pyqt import (
    QtNetwork, QWebChannel, QWebEnginePage, QWebEngineView, qt_version_info)
from photini.widgets import ComboBox, DoubleSpinBox, LatLongDisplay


logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate


class GeocoderBase(QtCore.QObject):
    interval = 5000
    cache_size = 200
    cache_ttl = 30 * 24 * 3600

    def __init__(self, *args, **kwds):
        super(GeocoderBase, self).__init__(*args, **kwds)
        self.app = QtWidgets.QApplication.instance()
        self.block_timer = QtCore.QTimer(self)
        self.block_timer.setInterval(self.interval)
        self.block_timer.setSingleShot(True)
        if self.cache_size:
            self.cache_file = os.path.join(
                appdirs.user_cache_dir('photini'),
                self.__class__.__name__ + '.pkl')
            try:
                with open(self.cache_file, 'rb') as f:
                    self.query_cache = pickle.load(f)
            except (AttributeError, FileNotFoundError):
                self.query_cache = cachetools.TTLCache(
                    self.cache_size, self.cache_ttl)
            logger.debug('cache %s has %d entries',
                         self.cache_file, self.query_cache.currsize)
            self.app.aboutToQuit.connect(self.save_cache)
        else:
            self.query_cache = None

    @QtSlot()
    @catch_all
    def save_cache(self):
        if self.query_cache:
            os.makedirs(os.path.dirname(self.cache_file), exist_ok=True)
            with open(self.cache_file, 'wb') as f:
                pickle.dump(self.query_cache, f)

    def rate_limit(self):
        while self.block_timer.isActive():
            self.app.processEvents()
        self.block_timer.start()

    def cached_query(self, params, *args):
        cache_key = ','.join(sorted([':'.join(x) for x in params.items()]))
        if cache_key in self.query_cache:
            return self.query_cache[cache_key]
        results = self.query(params, *args)
        self.query_cache[cache_key] = results
        return results


class CallHandler(QtCore.QObject):
    # Simple object (with no attributes) for JavaScript to send signals
    # to and hence invoke the methods JavaScript wants to call.
    @QtSlot(int, str)
    def log(self, level, message):
        logger.log(level, message)

    @QtSlot()
    def initialize_finished(self):
        self.parent().initialize_finished()

    @QtSlot('QVariant')
    def new_status(self, status):
        self.parent().new_status(status)

    @QtSlot(int)
    def marker_click(self, marker_id):
        self.parent().marker_click(marker_id)

    @QtSlot(float, float)
    def marker_drag(self, lat, lng):
        self.parent().marker_drag(lat, lng)

    @QtSlot(float, float, int)
    def marker_drag_end(self, lat, lng, marker_id):
        self.parent().marker_drag_end(lat, lng, marker_id)

    @QtSlot(float, float)
    def marker_drop(self, lat, lng):
        self.parent().marker_drop(lat, lng)


class MapWebPage(QWebEnginePage):
    def __init__(self, *args, call_handler=None, transient=False, **kwds):
        super(MapWebPage, self).__init__(*args, **kwds)
        self.call_handler = call_handler
        self.transient = transient
        if self.call_handler:
            self.web_channel = QWebChannel(parent=self)
            self.setWebChannel(self.web_channel)
            self.web_channel.registerObject('python', self.call_handler)
        self.profile().setCachePath(
            os.path.join(appdirs.user_cache_dir('photini'), 'WebEngine'))

    @catch_all
    def acceptNavigationRequest(self, url, type_, isMainFrame):
        if type_ != self.NavigationType.NavigationTypeLinkClicked:
            return super(MapWebPage, self).acceptNavigationRequest(
                url, type_, isMainFrame)
        if url.isLocalFile():
            url.setScheme('http')
        QtGui.QDesktopServices.openUrl(url)
        if self.transient:
            # delete temporary child created by createWindow
            self.deleteLater()
        return False

    @catch_all
    def createWindow(self, type_):
        return MapWebPage(transient=True, parent=self)

    @catch_all
    def javaScriptConsoleMessage(self, level, msg, line, source):
        logger.log(
            logging.INFO + (level * 10), '%s line %d: %s', source, line, msg)


class MapWebView(QWebEngineView):
    drop_text = QtSignal(int, int, str)

    def __init__(self, call_handler, *args, **kwds):
        super(MapWebView, self).__init__(*args, **kwds)
        self.setPage(MapWebPage(call_handler=call_handler, parent=self))
        settings = self.settings()
        settings.setAttribute(
            settings.WebAttribute.Accelerated2dCanvasEnabled, False)
        settings.setAttribute(
            settings.WebAttribute.LocalContentCanAccessRemoteUrls, True)
        settings.setAttribute(
            settings.WebAttribute.LocalContentCanAccessFileUrls, True)

    @catch_all
    def dragEnterEvent(self, event):
        if not event.mimeData().hasFormat(DRAG_MIMETYPE):
            return super(MapWebView, self).dragEnterEvent(event)
        event.acceptProposedAction()

    @catch_all
    def dragMoveEvent(self, event):
        if not event.mimeData().hasFormat(DRAG_MIMETYPE):
            return super(MapWebView, self).dragMoveEvent(event)
        event.acceptProposedAction()

    @catch_all
    def dropEvent(self, event):
        if not event.mimeData().hasFormat(DRAG_MIMETYPE):
            return super(MapWebView, self).dropEvent(event)
        text = event.mimeData().data(DRAG_MIMETYPE).data().decode('utf-8')
        if text:
            if qt_version_info < (6, 0):
                pos = event.pos()
            else:
                pos = event.position().toPoint()
            self.drop_text.emit(pos.x(), pos.y(), text)
            event.acceptProposedAction()


class PhotiniMap(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super(PhotiniMap, self).__init__(parent)
        self.app = QtWidgets.QApplication.instance()
        name = self.__module__.split('.')[-1]
        self.script_dir = pkg_resources.resource_filename(
            'photini', 'data/' + name + '/')
        self.drag_icon = QtGui.QPixmap(
            os.path.join(self.script_dir, '../map_pin_grey.png'))
        self.drag_hotspot = 11, 35
        self.search_string = None
        self.map_loaded = 0     # not loaded
        self.marker_info = {}
        self.map_status = {}
        self.dropped_images = []
        self.geocoder = self.get_geocoder()
        self.gpx_ids = []
        self.widgets = {}
        self.setLayout(QtWidgets.QHBoxLayout())
        ## left side
        left_side = QtWidgets.QGridLayout()
        # latitude & longitude
        self.widgets['latlon'] = LatLongDisplay()
        left_side.addWidget(self.widgets['latlon'].label, 0, 0)
        self.widgets['latlon'].new_value.connect(self.new_latlon)
        left_side.addWidget(self.widgets['latlon'], 0, 1)
        # altitude
        label = QtWidgets.QLabel(translate('PhotiniMap', 'Altitude'))
        label.setAlignment(Qt.AlignmentFlag.AlignRight)
        left_side.addWidget(label, 1, 0)
        self.widgets['alt'] = DoubleSpinBox('exif:GPSAltitude')
        self.widgets['alt'].set_suffix(
            translate('PhotiniMap', ' m', 'metres altitude'))
        self.widgets['alt'].new_value.connect(self.new_value)
        left_side.addWidget(self.widgets['alt'], 1, 1)
        if hasattr(self.geocoder, 'get_altitude'):
            self.widgets['get_altitude'] = QtWidgets.QPushButton(
                translate('PhotiniMap', 'Get altitude from map'))
            self.widgets['get_altitude'].clicked.connect(self.get_altitude)
            left_side.addWidget(self.widgets['get_altitude'], 2, 1)
        else:
            self.widgets['get_altitude'] = None
        # search
        label = QtWidgets.QLabel(translate('PhotiniMap', 'Search'))
        label.setAlignment(Qt.AlignmentFlag.AlignRight)
        left_side.addWidget(label, 3, 0)
        self.widgets['search'] = ComboBox()
        self.widgets['search'].setEditable(True)
        self.widgets['search'].setInsertPolicy(ComboBox.InsertPolicy.NoInsert)
        self.widgets['search'].lineEdit().setPlaceholderText(
            translate('PhotiniMap', '<new search>'))
        self.widgets['search'].lineEdit().returnPressed.connect(self.search)
        self.widgets['search'].activated.connect(self.goto_search_result)
        self.clear_search()
        self.widgets['search'].setEnabled(False)
        left_side.addWidget(self.widgets['search'], 3, 1)
        # search terms and conditions
        for n, widget in enumerate(self.geocoder.search_terms()):
            left_side.addWidget(widget, n+4, 0, 1, 2)
        left_side.setColumnStretch(1, 1)
        left_side.setRowStretch(7, 1)
        # GPX importer
        if self.app.gpx_importer:
            button = QtWidgets.QPushButton(
                translate('PhotiniMap', 'Load GPX file'))
            button.clicked.connect(self.load_gpx)
            left_side.addWidget(button, 8, 1)
            self.widgets['set_from_gpx'] = QtWidgets.QPushButton(
                translate('PhotiniMap', 'Set coords from GPX'))
            self.widgets['set_from_gpx'].setEnabled(False)
            self.widgets['set_from_gpx'].clicked.connect(self.set_from_gpx)
            left_side.addWidget(self.widgets['set_from_gpx'], 9, 1)
            self.widgets['clear_gpx'] = QtWidgets.QPushButton(
                translate('PhotiniMap', 'Remove GPX data'))
            self.widgets['clear_gpx'].setEnabled(False)
            self.widgets['clear_gpx'].clicked.connect(self.clear_gpx)
            left_side.addWidget(self.widgets['clear_gpx'], 10, 1)
        self.layout().addLayout(left_side)
        # map
        # create handler for calls from JavaScript
        self.call_handler = CallHandler(parent=self)
        self.widgets['map'] = MapWebView(self.call_handler)
        self.widgets['map'].drop_text.connect(self.drop_text)
        self.widgets['map'].setAcceptDrops(False)
        self.layout().addWidget(self.widgets['map'])
        self.layout().setStretch(1, 1)
        # other init
        self.app.image_list.image_list_changed.connect(self.image_list_changed)

    @QtSlot()
    @catch_all
    def image_list_changed(self):
        if not self.isVisible():
            return
        # add or remove markers
        self.redraw_markers()

    @QtSlot()
    @catch_all
    def initialise(self):
        page = '''<!DOCTYPE html>
<html>
  <head>
    <meta charset="utf-8" />
    <meta name="viewport" content="initial-scale=1.0, user-scalable=no" />
    <style type="text/css">
      html, body {{ height: 100%; margin: 0; padding: 0 }}
      #mapDiv {{ position: relative; width: 100%; height: 100% }}
    </style>
{initialize}
{head}
    <script type="text/javascript" src="script.js"></script>
  </head>
  <body ondragstart="return false">
    <div id="mapDiv"></div>
  </body>
</html>'''
        lat, lng = self.app.config_store.get('map', 'centre', (51.0, 0.0))
        zoom = int(self.app.config_store.get('map', 'zoom', 11))
        initialize = '''    <script type="text/javascript"
      src="qrc:///qtwebchannel/qwebchannel.js">
    </script>
    <script type="text/javascript">
      var python;
      function initialize()
      {{
          new QWebChannel(qt.webChannelTransport, doLoadMap);
      }}
      function doLoadMap(channel)
      {{
          python = channel.objects.python;
          loadMap({lat}, {lng}, {zoom});
      }}
    </script>'''
        initialize = initialize.format(lat=lat, lng=lng, zoom=zoom)
        page = page.format(initialize=initialize, head=self.get_head())
        QtWidgets.QApplication.setOverrideCursor(Qt.CursorShape.WaitCursor)
        self.widgets['map'].setHtml(
            page, QtCore.QUrl.fromLocalFile(self.script_dir))

    @catch_all
    def initialize_finished(self):
        QtWidgets.QApplication.restoreOverrideCursor()
        self.map_loaded = 2     # finished loading
        self.widgets['search'].setEnabled(True)
        self.widgets['map'].setAcceptDrops(True)
        self.new_selection(
            self.app.image_list.get_selected_images(), adjust_map=False)

    def refresh(self):
        self.app.image_list.set_drag_to_map(self.drag_icon, self.drag_hotspot)
        selection = self.app.image_list.get_selected_images()
        self.update_display(selection, adjust_map=False)
        if self.map_loaded < 1:
            self.map_loaded = 1     # started loading
            self.initialise()
            return
        if self.map_loaded < 2:
            return
        lat, lng = self.app.config_store.get('map', 'centre')
        zoom = int(self.app.config_store.get('map', 'zoom'))
        self.JavaScript('setView({!r},{!r},{:d})'.format(lat, lng, zoom))
        self.new_selection(selection, adjust_map=False)

    def do_not_close(self):
        return False

    @catch_all
    def new_status(self, status):
        self.map_status.update(status)
        for key in ('centre', 'zoom'):
            if key in status:
                self.app.config_store.set('map', key, self.map_status[key])

    @QtSlot(int, int, str)
    @catch_all
    def drop_text(self, x, y, text):
        self.dropped_images = eval(text)
        self.JavaScript('markerDrop({:d},{:d})'.format(x, y))

    @catch_all
    def marker_drop(self, lat, lng):
        value = {'exif:GPSLatitude': lat, 'exif:GPSLongitude': lng}
        images = [self.app.image_list.get_image(path)
                  for path in self.dropped_images]
        self.dropped_images = []
        self.new_value(value, images=images, adjust_map=False)

    @QtSlot(dict)
    @catch_all
    def new_latlon(self, value):
        self.new_value(value, adjust_map=True)

    @QtSlot(dict)
    @catch_all
    def new_value(self, value, images=[], adjust_map=False):
        images = images or self.app.image_list.get_selected_images()
        value['method'] = 'MANUAL'
        for image in images:
            gps = dict(image.metadata.gps_info)
            gps.update(value)
            image.metadata.gps_info = gps
        self.update_display(images, adjust_map=adjust_map)

    def update_display(self, images, adjust_map=True):
        self.redraw_markers()
        values = []
        for image in images:
            values.append(image.metadata.gps_info)
        self.widgets['latlon'].set_value_list(values)
        self.widgets['alt'].set_value_list(values)
        if self.widgets['get_altitude']:
            self.widgets['get_altitude'].setEnabled(
                bool(self.widgets['latlon'].get_value()))
        if adjust_map:
            self.see_selection(images)

    def see_selection(self, selected_images):
        locations = []
        # get locations of selected images
        for image in selected_images:
            gps = image.metadata.gps_info
            if not gps['exif:GPSLatitude']:
                continue
            location = [float(gps['exif:GPSLatitude']),
                        float(gps['exif:GPSLongitude'])]
            if location not in locations:
                locations.append(location)
        if not locations:
            # get locations of GPS track points around time of selected images
            for point in self.get_nearest_gps(selected_images):
                time_stamp, lat, lng = point
                location = [lat, lng]
                if location not in locations:
                    locations.append(location)
        # adjust map
        if locations:
            self.JavaScript('fitPoints({!r})'.format(locations))

    def new_selection(self, selection, adjust_map=True):
        if 'set_from_gpx' in self.widgets:
            self.widgets['set_from_gpx'].setEnabled(
                bool(selection) and bool(self.app.gpx_importer.display_points))
        self.redraw_gps_track(selection)
        self.update_display(selection, adjust_map=adjust_map)

    def redraw_markers(self):
        if self.map_loaded < 2:
            return
        for info in self.marker_info.values():
            info['images'] = []
        # assign images to existing markers or create new markers
        for image in self.app.image_list.get_images():
            gps = image.metadata.gps_info
            if not gps['exif:GPSLatitude']:
                continue
            location = [float(gps['exif:GPSLatitude']),
                        float(gps['exif:GPSLongitude'])]
            for info in self.marker_info.values():
                if info['location'] == location:
                    info['images'].append(image)
                    break
            else:
                for i in range(len(self.marker_info) + 2):
                    marker_id = i
                    if marker_id not in self.marker_info:
                        break
                self.marker_info[marker_id] = {
                    'images'  : [image],
                    'location': location,
                    'selected': image.selected,
                    }
                self.JavaScript('addMarker({:d},{!r},{!r},{:d})'.format(
                    marker_id, location[0], location[1], image.selected))
        # delete redundant markers and enable markers with selected images
        for marker_id in list(self.marker_info.keys()):
            info = self.marker_info[marker_id]
            if not info['images']:
                self.JavaScript('delMarker({:d})'.format(marker_id))
                del self.marker_info[marker_id]
            elif info['selected'] != any([x.selected for x in info['images']]):
                info['selected'] = not info['selected']
                self.JavaScript(
                    'enableMarker({:d},{:d})'.format(marker_id, info['selected']))

    def redraw_gps_track(self, selected_images=None):
        if self.map_loaded < 2:
            return
        # update GPX track markers
        if not self.app.gpx_importer:
            return
        if not self.app.gpx_importer.display_points:
            self.JavaScript('clearGPS()')
            self.gpx_ids = []
            return
        # add any new points
        new_points = []
        for time_stamp, lat, lng in self.app.gpx_importer.display_points:
            marker_id = time_stamp.isoformat()
            if marker_id not in self.gpx_ids:
                self.gpx_ids.append(marker_id)
                new_points.append([lat, lng, marker_id])
        if new_points:
            self.JavaScript('plotGPS({!r})'.format(new_points))
        # highlight points near selected picture
        if selected_images is None:
            selected_images = self.app.image_list.get_selected_images()
        selected_ids = [x[0].isoformat()
                        for x in self.get_nearest_gps(selected_images)]
        self.JavaScript('enableGPS({!r})'.format(selected_ids))

    def get_nearest_gps(self, selected_images):
        if not (selected_images and self.app.gpx_importer):
            return []
        date_taken = selected_images[0].metadata.date_taken
        for image in selected_images[1:]:
            if image.metadata.date_taken != date_taken:
                return []
        if not date_taken:
            return []
        return self.app.gpx_importer.nearest(date_taken.to_utc())

    @QtSlot()
    @catch_all
    def load_gpx(self):
        new_points = self.app.gpx_importer.import_file()
        if not new_points:
            return
        self.redraw_gps_track()
        # adjust map to show points just loaded
        latlngs = []
        for p in new_points:
            time_stamp, lat, lng = p
            latlngs.append([lat, lng])
        self.JavaScript('fitPoints({!r})'.format(latlngs))
        self.widgets['set_from_gpx'].setEnabled(
            bool(self.app.image_list.get_selected_images())
            and bool(self.app.gpx_importer.display_points))
        self.widgets['clear_gpx'].setEnabled(
            bool(self.app.gpx_importer.display_points))

    @QtSlot()
    @catch_all
    def set_from_gpx(self):
        selected_images = self.app.image_list.get_selected_images()
        set_altitude = self.app.config_store.get('map', 'gpx_altitude', True)
        changed = False
        for image in selected_images:
            if not image.metadata.date_taken:
                continue
            utc_time = image.metadata.date_taken.to_utc()
            utc_time = utc_time.replace(tzinfo=timezone.utc)
            candidates = self.app.gpx_importer.get_locations_at(utc_time)
            if not candidates:
                continue
            nearest = candidates[0]
            for c in candidates[1:]:
                if abs(c.time - utc_time) < abs(nearest.time - utc_time):
                    nearest = c
            gps = dict(image.metadata.gps_info)
            gps['exif:GPSLatitude'] = nearest.latitude
            gps['exif:GPSLongitude'] = nearest.longitude
            if set_altitude and nearest.elevation is not None:
                gps['exif:GPSAltitude'] = round(nearest.elevation, 1)
            else:
                gps['exif:GPSAltitude'] = None
            gps['method'] = 'GPS'
            image.metadata.gps_info = gps
            changed = True
        if changed:
            self.update_display(selected_images)

    @QtSlot()
    @catch_all
    def clear_gpx(self):
        self.app.gpx_importer.clear_data()
        self.redraw_gps_track()
        self.widgets['set_from_gpx'].setEnabled(False)
        self.widgets['clear_gpx'].setEnabled(False)

    @QtSlot()
    @catch_all
    def get_altitude(self):
        images = self.app.image_list.get_selected_images()
        altitude = self.geocoder.get_altitude(
            self.widgets['latlon'].get_value())
        if altitude is not None:
            self.new_value(
                {'exif:GPSAltitude': round(altitude, 1)}, images=images)

    @QtSlot()
    @catch_all
    def search(self, search_string=None, bounded=True):
        if not search_string:
            search_string = self.widgets['search'].lineEdit().text()
            self.widgets['search'].clearEditText()
        if not search_string:
            return
        self.search_string = search_string
        self.clear_search()
        if bounded:
            bounds = self.map_status['bounds']
        else:
            bounds = None
        for result in self.geocoder.search(search_string, bounds=bounds):
            north, east, south, west, name = result
            self.widgets['search'].addItem(name, (north, east, south, west))
        self.widgets['search'].set_dropdown_width()
        self.widgets['search'].showPopup()

    def clear_search(self):
        self.widgets['search'].clear()
        self.widgets['search'].addItem('')
        if self.search_string:
            self.widgets['search'].addItem(
                translate('PhotiniMap', '<widen search>'), 'widen')
            self.widgets['search'].addItem(
                translate('PhotiniMap', '<repeat search>'), 'repeat')

    @QtSlot(int)
    @catch_all
    def goto_search_result(self, idx):
        self.widgets['search'].setCurrentIndex(0)
        self.widgets['search'].clearFocus()
        if idx == 0:
            return
        data = self.widgets['search'].itemData(idx)
        if data is None:
            return
        if data == 'widen':
            # widen search
            self.search(search_string=self.search_string, bounded=False)
        elif data == 'repeat':
            # repeat search
            self.search(search_string=self.search_string)
        elif data[-1] is None:
            self.JavaScript('setView({},{},{})'.format(
                data[0], data[1], self.map_status['zoom']))
        else:
            self.JavaScript('adjustBounds({},{},{},{})'.format(*data))

    @catch_all
    def marker_click(self, marker_id):
        self.app.image_list.select_images(self.marker_info[marker_id]['images'])

    @catch_all
    def marker_drag(self, lat, lng):
        self.widgets['latlon'].set_value((lat, lng))

    @catch_all
    def marker_drag_end(self, lat, lng, marker_id):
        info = self.marker_info[marker_id]
        for image in info['images']:
            gps = dict(image.metadata.gps_info)
            gps['exif:GPSLatitude'] = lat
            gps['exif:GPSLongitude'] = lng
            gps['method'] = 'MANUAL'
            image.metadata.gps_info = gps
        info['location'] = [float(image.metadata.gps_info['exif:GPSLatitude']),
                            float(image.metadata.gps_info['exif:GPSLongitude'])]
        self.widgets['latlon'].set_value_list(
            [info['images'][0].metadata.gps_info])

    def JavaScript(self, command):
        if self.map_loaded >= 2:
            self.widgets['map'].page().runJavaScript(command)



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2015-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from collections import namedtuple
from contextlib import contextmanager
from functools import wraps
import importlib
import logging
import os
import sys

from photini.configstore import BaseConfigStore

logger = logging.getLogger(__name__)

# allow a "standard" list of objects to be imported with *
__all__ = (
    'Busy', 'catch_all', 'DisableWidget', 'execute', 'flag_to_int',
    'FormLayout', 'multiple', 'multiple_values', 'Qt', 'QtCore', 'QtGui',
    'QtGui2', 'QtSignal', 'QtSlot', 'QtWidgets', 'scale_font', 'UnBusy',
    'width_for_text', 'wrap_text'
    )

# temporarily open config file to get any over-rides
config = BaseConfigStore('editor')
config.delete('pyqt', 'using_pyqt5')
using_pyside = config.get('pyqt', 'using_pyside2')
config.delete('pyqt', 'using_pyside2')
config.delete('pyqt', 'using_qtwebengine')
qt_lib = config.get('pyqt', 'qt_lib', 'auto')
if qt_lib == 'auto' and isinstance(using_pyside, bool):
    # copy old config
    qt_lib = ('PyQt5', 'PySide2')[using_pyside]
    config.set('pyqt', 'qt_lib', qt_lib)
qt_scale_factor = config.get('pyqt', 'scale_factor', 1)
if qt_scale_factor != 1:
    os.environ['QT_SCALE_FACTOR'] = str(qt_scale_factor)

def import_PySide6():
    global QtCore, QtGui, QtNetwork, QtWidgets
    global Qt, QtSignal, QtSlot, PySide_version
    global QtGui2
    global QWebChannel, QWebEngineView, QWebEnginePage
    from PySide6 import QtCore, QtGui, QtNetwork, QtWidgets
    from PySide6.QtCore import Qt
    from PySide6.QtCore import Signal as QtSignal
    from PySide6.QtCore import Slot as QtSlot
    from PySide6 import __version__ as PySide_version
    QtGui2 = QtGui
    from PySide6.QtWebChannel import QWebChannel
    from PySide6.QtWebEngineWidgets import QWebEngineView
    from PySide6.QtWebEngineCore import QWebEnginePage

def import_PySide2():
    global QtCore, QtGui, QtNetwork, QtWidgets
    global Qt, QtSignal, QtSlot, PySide_version
    global QtGui2
    global QWebChannel, QWebEngineView, QWebEnginePage
    from PySide2 import QtCore, QtGui, QtNetwork, QtWidgets
    from PySide2.QtCore import Qt
    from PySide2.QtCore import Signal as QtSignal
    from PySide2.QtCore import Slot as QtSlot
    from PySide2 import __version__ as PySide_version
    QtGui2 = QtWidgets
    from PySide2.QtWebChannel import QWebChannel
    from PySide2.QtWebEngineWidgets import QWebEnginePage, QWebEngineView

def import_PyQt6():
    global QtCore, QtGui, QtNetwork, QtWidgets
    global Qt, QtSignal, QtSlot
    global QtGui2
    global QWebChannel, QWebEngineView, QWebEnginePage
    from PyQt6 import QtCore, QtGui, QtNetwork, QtWidgets
    from PyQt6.QtCore import Qt
    from PyQt6.QtCore import pyqtSignal as QtSignal
    from PyQt6.QtCore import pyqtSlot as QtSlot
    QtGui2 = QtGui
    from PyQt6.QtWebChannel import QWebChannel
    from PyQt6.QtWebEngineWidgets import QWebEngineView
    from PyQt6.QtWebEngineCore import QWebEnginePage

def import_PyQt5():
    global QtCore, QtGui, QtNetwork, QtWidgets
    global Qt, QtSignal, QtSlot
    global QtGui2
    global QWebChannel, QWebEngineView, QWebEnginePage
    from PyQt5 import QtCore, QtGui, QtNetwork, QtWidgets
    from PyQt5.QtCore import Qt
    from PyQt5.QtCore import pyqtSignal as QtSignal
    from PyQt5.QtCore import pyqtSlot as QtSlot
    QtGui2 = QtWidgets
    from PyQt5.QtWebChannel import QWebChannel
    from PyQt5.QtWebEngineWidgets import QWebEnginePage, QWebEngineView

# choose Qt package
if qt_lib == 'auto':
    _libs = ('PyQt5', 'PyQt6', 'PySide2', 'PySide6')
else:
    _libs = (qt_lib,)
_error = ['']
for key in _libs:
    try:
        {'PyQt5': import_PyQt5,
         'PyQt6': import_PyQt6,
         'PySide2': import_PySide2,
         'PySide6': import_PySide6,
         }[key]()
        qt_lib = key
        break
    except KeyError:
        raise ImportError('Unknown Qt library ' + key) from None
    except ImportError as ex:
        if qt_lib != 'auto':
            raise ex
        _error.append('{} import failed: {}'.format(key, str(ex)))
else:
    raise ImportError('\n'.join(_error))
using_pyside = 'PySide' in qt_lib

style = config.get('pyqt', 'style')
if style:
    QtWidgets.QApplication.setStyle(style)
config.save()
del config, style

if using_pyside:
    qt_version_info = QtCore.__version_info__
    qt_version = '{} {}, Qt {}'.format(
        qt_lib, PySide_version, QtCore.__version__)
else:
    qt_version_info = namedtuple(
        'qt_version_info', ('major', 'minor', 'micro'))._make(
            map(int, QtCore.QT_VERSION_STR.split('.')))
    qt_version = 'PyQt {}, Qt {}'.format(
        QtCore.PYQT_VERSION_STR, QtCore.QT_VERSION_STR)
    pyqt_version_info = namedtuple(
        'pyqt_version_info', ('major', 'minor', 'micro'))._make(
            map(int, QtCore.PYQT_VERSION_STR.split('.')))
    if pyqt_version_info < (5, 11):
        raise ImportError(
            'PyQt version {}.{}.{} is less than 5.11'.format(*pyqt_version_info))

if qt_version_info < (5, 8):
    raise ImportError(
        'Qt version {}.{}.{} is less than 5.8'.format(*qt_version_info))

# increase image size that can be read
if qt_version_info >= (6, 0):
    if qt_version_info > (6, 3, 1) or using_pyside:
        QtGui.QImageReader.setAllocationLimit(0)
    else:
        os.environ['QT_IMAGEIO_MAXALLOC'] = '256'

# set network proxy
QtNetwork.QNetworkProxyFactory.setUseSystemConfiguration(True)

# workaround for Qt bug affecting QtWebEngine
# https://bugreports.qt.io/browse/QTBUG-67537
if sys.platform.startswith('linux') and qt_version_info < (5, 11, 0):
    import ctypes
    import ctypes.util
    ctypes.CDLL(ctypes.util.find_library('GL'), ctypes.RTLD_GLOBAL)

translate = QtCore.QCoreApplication.translate

# decorator for methods called by Qt that logs any exception raised
def catch_all(func):
    @wraps(func)
    def wrapper(*args, **kwds):
        try:
            return func(*args, **kwds)
        except Exception as ex:
            logger.exception(ex)
    return wrapper


def image_types_lower():
    result = [
        'jpeg', 'jpg', 'exv', 'cr2', 'crw', 'mrw', 'tiff', 'tif', 'dng',
        'nef', 'pef', 'arw', 'rw2', 'sr2', 'srw', 'orf', 'png', 'pgf',
        'raf', 'eps', 'gif', 'psd', 'tga', 'bmp', 'jp2', 'pnm',
        'cr3', 'heif', 'heic', 'avif'
        ]
    for fmt in QtGui.QImageReader.supportedImageFormats():
        ext = fmt.data().decode('utf-8').lower()
        if ext not in result:
            result.append(ext)
    for ext in ('ico', 'xcf'):
        if ext in result:
            result.remove(ext)
    return result

def image_types():
    lower = image_types_lower()
    return lower + [x.upper() for x in lower] + [x.title() for x in lower]

def video_types_lower():
    return ['3gp', 'avi', 'mp4', 'mpeg', 'mpg', 'mov', 'mts', 'qt', 'wmv']

def video_types():
    lower = video_types_lower()
    return lower + [x.upper() for x in lower] + [x.title() for x in lower]

def multiple():
    return translate('Widgets', '<multiple>')

def multiple_values():
    return translate('Widgets', '<multiple values>')

def set_symbol_font(widget):
    widget.setFont(QtGui.QFont('DejaVu Sans'))
    if widget.fontInfo().family().lower() != 'dejavu sans':
        # probably on Windows, try a different font
        widget.setFont(QtGui.QFont("Segoe UI Symbol"))

def scale_font(widget, scale):
    font = widget.font()
    size = font.pointSizeF()
    if size < 0:
        size = font.pixelSize()
        font.setPixelSize(((size * scale) + 50) // 100)
    else:
        font.setPointSizeF(size * scale / 100.0)
    widget.setFont(font)

def width_for_text(widget, text):
    rect = widget.fontMetrics().boundingRect(text)
    return rect.width()

def wrap_text(widget, text, lines):
    def sub_wrap(fm, text, lines, idx):
        width = fm.boundingRect(0, 0, 1, 1, 0, text).width()
        result = text
        while True:
            idx = text.find(' ', idx + 1)
            if idx < 0:
                break
            new_text = '\n'.join((text[:idx], text[idx+1:]))
            if lines > 2:
                new_text = sub_wrap(fm, new_text, lines-1, idx+1)
            new_width = fm.boundingRect(0, 0, 1, 1, 0, new_text).width()
            if width > new_width:
                width = new_width
                result = new_text
        return result
    return sub_wrap(widget.fontMetrics(), text, lines, 0)

def execute(widget, *arg, **kwds):
    if qt_lib == 'PySide2':
        return widget.exec_(*arg, **kwds)
    return widget.exec(*arg, **kwds)

def flag_to_int(flags):
    if hasattr(flags, 'value'):
        # recent versions of PyQt6 and PySide6 convert QEnum to Python enum
        return flags.value
    return int(flags)


@contextmanager
def Busy():
    QtWidgets.QApplication.setOverrideCursor(Qt.CursorShape.WaitCursor)
    try:
        yield
    finally:
        QtWidgets.QApplication.restoreOverrideCursor()


@contextmanager
def UnBusy():
    cursors = []
    while True:
        cursor = QtWidgets.QApplication.overrideCursor()
        if not cursor:
            break
        cursors.append(cursor.shape())
        QtWidgets.QApplication.restoreOverrideCursor()
    try:
        yield
    finally:
        while cursors:
            QtWidgets.QApplication.setOverrideCursor(cursors.pop())


@contextmanager
def DisableWidget(widget):
    widget.setEnabled(False)
    QtWidgets.QApplication.processEvents()
    try:
        yield
    finally:
        widget.setEnabled(True)


class FormLayout(QtWidgets.QFormLayout):
    def __init__(self, wrapped=False, **kwds):
        super(FormLayout, self).__init__(**kwds)
        if wrapped:
            self.setRowWrapPolicy(self.RowWrapPolicy.WrapAllRows)
        self.setFieldGrowthPolicy(self.FieldGrowthPolicy.AllNonFixedFieldsGrow)



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2022-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

import logging
import re

from photini.pyqt import *
from photini.pyqt import qt_version_info
from photini.types import MD_LangAlt

logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate


class WidgetMixin(object):
    new_value = QtSignal(dict)

    @QtSlot()
    @catch_all
    def emit_value(self):
        if not self.is_multiple():
            self.new_value.emit(self.get_value_dict())

    def get_value_dict(self):
        if self.is_multiple():
            return {}
        return {self._key: self.get_value()}

    def set_value_list(self, values):
        if not values:
            self.setEnabled(False)
            self.set_value(None)
            return
        self.setEnabled(True)
        choices = []
        for value_dict in values:
            value = None
            if self._key in value_dict:
                value = value_dict[self._key]
            if value not in choices:
                choices.append(value)
        if len(choices) > 1:
            self.set_multiple(choices=[x for x in choices if x])
        else:
            self.set_value(choices and choices[0])


class ComboBox(QtWidgets.QComboBox):
    def __init__(self, *args, **kwds):
        super(ComboBox, self).__init__(*args, **kwds)
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

    @catch_all
    def wheelEvent(self, event):
        if self.hasFocus():
            return super(ComboBox, self).wheelEvent(event)
        event.ignore()

    def set_dropdown_width(self):
        width = 0
        for idx in range(self.count()):
            width = max(width, width_for_text(self, self.itemText(idx) + 'xx'))
        margin = self.view().verticalScrollBar().sizeHint().width()
        self.view().setMinimumWidth(width + margin)


class Label(QtWidgets.QLabel):
    def __init__(self, *args, lines=1, layout=None, **kwds):
        super(Label, self).__init__(*args, **kwds)
        if lines == 1:
            return
        self.setText(wrap_text(self, self.text(), lines))
        if not layout:
            return
        # match text alignment to form layout
        align_h = (layout.labelAlignment()
                   & Qt.AlignmentFlag.AlignHorizontal_Mask)
        align_v = self.alignment() & Qt.AlignmentFlag.AlignVertical_Mask
        self.setAlignment(align_h | align_v)
        # Qt internally makes labels in form layouts 7/4 as tall, which
        # is too much for multi=line labels
        height = self.sizeHint().height() * ((lines * 4) + 3) // (lines * 4)
        self.setFixedHeight(height)


class PushButton(QtWidgets.QPushButton):
    def __init__(self, *args, lines=1, **kwds):
        super(PushButton, self).__init__(*args, **kwds)
        if lines != 1:
            self.setText(wrap_text(self, self.text(), lines))


class CompactButton(QtWidgets.QPushButton):
    def __init__(self, *args, **kwds):
        super(CompactButton, self).__init__(*args, **kwds)
        if QtWidgets.QApplication.style().objectName() in ('breeze',):
            self.setStyleSheet('padding: 2px;')
        scale_font(self, 80)


class DropDownSelector(ComboBox, WidgetMixin):
    def __init__(self, key, values=[], default=None,
                 with_multiple=True, extendable=False, ordered=False):
        super(DropDownSelector, self).__init__()
        self._key = key
        self._with_multiple = with_multiple
        self._extendable = extendable
        self._ordered = ordered
        self._old_idx = 0
        self.setSizeAdjustPolicy(
            self.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)
        self.set_values(values, default=default)
        self.currentIndexChanged.connect(self._index_changed)

    def add_item(self, text, data, adjust_width=True):
        idx = self.findText(text)
        if idx >= 0:
            return idx
        idx = self._last_idx()
        if self._ordered:
            for n in range(idx):
                if self.itemText(n).lower() > text.lower():
                    idx = n
                    break
        blocked = self.blockSignals(True)
        self.insertItem(idx, text, data)
        self.blockSignals(blocked)
        if adjust_width:
            self.set_dropdown_width()
        return idx

    def remove_item(self, data, adjust_width=True):
        idx = self.findData(data)
        if idx < 0:
            return
        blocked = self.blockSignals(True)
        self.removeItem(idx)
        self.blockSignals(blocked)

    def set_values(self, values, default=None):
        blocked = self.blockSignals(True)
        self.clear()
        if self._extendable:
            self.addItem(translate('Widgets', '<new>'))
        if self._with_multiple:
            self.addItem(multiple_values())
            self.setItemData(self.count() - 1, '', Qt.ItemDataRole.UserRole - 1)
        for text, data in values:
            self.add_item(text, data, adjust_width=False)
        self.set_dropdown_width()
        self.blockSignals(blocked)
        if default is not None:
            self.set_value(default)

    def get_value(self):
        return self.itemData(self.currentIndex())

    def set_value(self, value):
        self._old_idx = self.findData(value)
        if self._old_idx < 0:
            self._old_idx = self.add_item(self.data_to_text(value), value)
        blocked = self.blockSignals(True)
        self.setCurrentIndex(self._old_idx)
        self.blockSignals(blocked)

    def data_to_text(self, data):
        return str(data)

    def is_multiple(self):
        return self._with_multiple and self.currentIndex() == self.count() - 1

    def set_multiple(self, choices=[]):
        if self._with_multiple:
            blocked = self.blockSignals(True)
            self._old_idx = self.count() - 1
            self.setCurrentIndex(self._old_idx)
            self.blockSignals(blocked)

    @catch_all
    def findData(self, data):
        # Qt's findData only works with simple types
        for n in range(self._last_idx()):
            if self.itemData(n) == data:
                return n
        return -1

    @QtSlot(int)
    @catch_all
    def _index_changed(self, idx):
        if idx < self._last_idx():
            # normal item selection
            self._old_idx = idx
            self.emit_value()
            return
        # user must have clicked '<new>'
        blocked = self.blockSignals(True)
        self.setCurrentIndex(self._old_idx)
        self.blockSignals(blocked)
        text, data = self.define_new_value()
        if not text:
            return
        blocked = self.blockSignals(True)
        self._old_idx = self.findText(text)
        if self._old_idx >= 0:
            # update existing item
            self.setItemData(self._old_idx, data)
        else:
            # add new item
            self._old_idx = self.add_item(text, data)
        self.setCurrentIndex(self._old_idx)
        self.blockSignals(blocked)
        self.emit_value()

    def _last_idx(self):
        idx = self.count()
        if self._extendable:
            idx -= 1
        if self._with_multiple:
            idx -= 1
        return idx


class TextHighlighter(QtGui.QSyntaxHighlighter):
    def __init__(self, spelling, length, length_always, multi_string, parent):
        super(TextHighlighter, self).__init__(parent)
        self.config_store = QtWidgets.QApplication.instance().config_store
        if spelling:
            self.spell_check = QtWidgets.QApplication.instance().spell_check
            self.spell_check.new_dict.connect(self.rehighlight)
            self.spell_formatter = QtGui.QTextCharFormat()
            self.spell_formatter.setUnderlineColor(Qt.GlobalColor.red)
            self.spell_formatter.setUnderlineStyle(
                QtGui.QTextCharFormat.UnderlineStyle.SpellCheckUnderline)
            self.find_words = self.spell_check.find_words
            self.suggest = self.spell_check.suggest
        else:
            self.spell_check = None
        if length:
            self.length_check = length
            self.length_always = length_always
            self.length_formatter = QtGui.QTextCharFormat()
            self.length_formatter.setUnderlineColor(Qt.GlobalColor.blue)
            self.length_formatter.setUnderlineStyle(
                QtGui.QTextCharFormat.UnderlineStyle.SingleUnderline)
            if multi_string:
                # treat each keyword separately
                self.pattern = re.compile(r'\s*(.+?)(;|$)')
            else:
                # treat the entire block as one
                self.pattern = re.compile(r'(.+)')
        else:
            self.length_check = None

    @catch_all
    def highlightBlock(self, text):
        if not text:
            if self.length_check:
                self.setCurrentBlockState(self.previousBlockState())
            return
        if self.length_check:
            length_warning = self.length_always or self.config_store.get(
                'files', 'length_warning', True)
            if length_warning:
                consumed = max(self.previousBlockState(), 0)
                max_len = max(self.length_check - consumed, 0)
                for match in self.pattern.finditer(text):
                    start = match.start(1)
                    end = match.end(1)
                    truncated = text[start:end].encode('utf-8')
                    consumed += len(truncated)
                    truncated = truncated[:max_len]
                    start += len(truncated.decode('utf-8', errors='ignore'))
                    if start < end:
                        self.setFormat(start, end - start,
                                       self.length_formatter)
                self.setCurrentBlockState(max(consumed, 0))
        if self.spell_check:
            for word, start, end in self.find_words(text):
                if not self.spell_check.check(word):
                    self.setFormat(start, end - start, self.spell_formatter)


class MultiLineEdit(QtWidgets.QPlainTextEdit, WidgetMixin):
    def __init__(self, key, *arg, spell_check=False, length_check=None,
                 multi_string=False, length_always=False, min_width=None, **kw):
        super(MultiLineEdit, self).__init__(*arg, **kw)
        if min_width:
            self.setMinimumWidth(width_for_text(self, 'x' * min_width))
        if self.isRightToLeft():
            self.set_text_alignment(Qt.AlignmentFlag.AlignRight)
        self._key = key
        self.multiple_values = multiple_values()
        self.setTabChangesFocus(True)
        self._is_multiple = False
        self.spell_check = spell_check
        self.highlighter = TextHighlighter(
            spell_check, length_check, length_always, multi_string,
            self.document())

    @catch_all
    def focusOutEvent(self, event):
        self.emit_value()
        super(MultiLineEdit, self).focusOutEvent(event)

    @catch_all
    def keyPressEvent(self, event):
        if self._is_multiple:
            self._is_multiple = False
            self.setPlaceholderText('')
        super(MultiLineEdit, self).keyPressEvent(event)

    @catch_all
    def contextMenuEvent(self, event):
        menu = self.createStandardContextMenu()
        suggestion_group = QtGui2.QActionGroup(menu)
        if self._is_multiple:
            if self.choices:
                sep = menu.insertSeparator(menu.actions()[0])
                fm = menu.fontMetrics()
                for suggestion in self.choices:
                    label = str(suggestion).replace('\n', ' ')
                    label = fm.elidedText(
                        label, Qt.TextElideMode.ElideMiddle, self.width())
                    action = QtGui2.QAction(label, suggestion_group)
                    action.setData(suggestion)
                    menu.insertAction(sep, action)
        elif self.spell_check:
            cursor = self.cursorForPosition(event.pos())
            block_pos = cursor.block().position()
            for word, start, end in self.highlighter.find_words(
                                                        cursor.block().text()):
                if start > cursor.positionInBlock():
                    break
                if end <= cursor.positionInBlock():
                    continue
                cursor.setPosition(block_pos + start)
                cursor.setPosition(block_pos + end, cursor.MoveMode.KeepAnchor)
                break
            suggestions = self.highlighter.suggest(cursor.selectedText())
            if suggestions:
                sep = menu.insertSeparator(menu.actions()[0])
                for suggestion in suggestions:
                    action = QtGui2.QAction(suggestion, suggestion_group)
                    menu.insertAction(sep, action)
        action = execute(menu, event.globalPos())
        if action and action.actionGroup() == suggestion_group:
            if self._is_multiple:
                self.set_value(action.data())
                self.emit_value()
            else:
                cursor.setPosition(block_pos + start)
                cursor.setPosition(block_pos + end, cursor.MoveMode.KeepAnchor)
                cursor.insertText(action.iconText())

    def set_height(self, rows):
        height = QtWidgets.QLineEdit().sizeHint().height()
        height += (rows - 1) * self.fontMetrics().lineSpacing()
        self.setMaximumHeight(height)

    def set_value(self, value):
        if self._is_multiple:
            self._is_multiple = False
            self.setPlaceholderText('')
        if not value:
            self.clear()
        else:
            self.setPlainText(str(value))

    def get_value(self):
        if qt_version_info < (5, 9):
            return self.toPlainText()
        value = self.document().toRawText()
        value = value.replace('\u2029', '\n')
        return value

    def set_multiple(self, choices=[]):
        self._is_multiple = True
        self.choices = list(choices)
        self.setPlaceholderText(self.multiple_values)
        self.clear()

    def is_multiple(self):
        return self._is_multiple and not bool(self.get_value())

    def set_text_alignment(self, alignment):
        options = self.document().defaultTextOption()
        options.setAlignment(alignment)
        self.document().setDefaultTextOption(options)


class SingleLineEdit(MultiLineEdit):
    def __init__(self, *arg, **kw):
        super(SingleLineEdit, self).__init__(*arg, **kw)
        self.setFixedHeight(QtWidgets.QLineEdit().sizeHint().height())
        self.setLineWrapMode(QtWidgets.QPlainTextEdit.LineWrapMode.NoWrap)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

    @catch_all
    def keyPressEvent(self, event):
        if event.key() == Qt.Key.Key_Return:
            event.ignore()
            return
        super(SingleLineEdit, self).keyPressEvent(event)

    @catch_all
    def insertFromMimeData(self, source):
        self.insertPlainText(source.text().replace('\n', ' '))


class MultiStringEdit(SingleLineEdit):
    def set_value(self, value):
        if isinstance(value, (list, tuple)):
            value = '; '.join(value)
        super(MultiStringEdit, self).set_value(value)

    def get_value(self):
        value = super(MultiStringEdit, self).get_value().split(';')
        value = [x.strip() for x in value]
        return [x for x in value if x]


class Slider(QtWidgets.QSlider, WidgetMixin):
    def __init__(self, key, *arg, **kw):
        super(Slider, self).__init__(*arg, **kw)
        self._key = key
        self._is_multiple = False
        self.sliderPressed.connect(self.slider_pressed)

    @catch_all
    def focusOutEvent(self, event):
        self.emit_value()
        super(Slider, self).focusOutEvent(event)

    @QtSlot()
    @catch_all
    def slider_pressed(self):
        self._is_multiple = False

    def get_value(self):
        return self.value()

    def set_value(self, value):
        self._is_multiple = False
        if value is not None:
            self.setValue(value)

    def set_multiple(self, choices=[]):
        self._is_multiple = True
        value = self.value()
        for choice in choices:
            if choice is not None:
                value = max(value, choice)
        self.setValue(value)

    def is_multiple(self):
        return self._is_multiple


class StartStopButton(QtWidgets.QPushButton):
    click_start = QtSignal()
    click_stop = QtSignal()

    def __init__(self, start_text, stop_text, lines=1):
        super(StartStopButton, self).__init__()
        if lines != 1:
            start_text = wrap_text(self, start_text, lines)
            stop_text = wrap_text(self, stop_text, lines)
        self.start_text = start_text
        self.stop_text = stop_text
        self.checked = False
        self.clicked.connect(self.do_clicked)
        # get a size big enough for either text
        self.setText(self.stop_text)
        stop_size = super(StartStopButton, self).sizeHint()
        self.setText(self.start_text)
        start_size = super(StartStopButton, self).sizeHint()
        self.minimum_size = stop_size.expandedTo(start_size)

    @catch_all
    def sizeHint(self):
        return self.minimum_size

    def is_checked(self):
        return self.checked

    def set_checked(self, value):
        self.checked = value
        if self.checked:
            self.setText(self.stop_text)
        else:
            self.setText(self.start_text)

    @QtSlot()
    def do_clicked(self):
        if self.checked:
            self.click_stop.emit()
        else:
            self.click_start.emit()


class LangAltWidget(QtWidgets.QWidget, WidgetMixin):
    def __init__(self, key, multi_line=True, label=None, **kw):
        super(LangAltWidget, self).__init__()
        layout = QtWidgets.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(
            (layout.alignment() & Qt.AlignmentFlag.AlignHorizontal_Mask)
            | Qt.AlignmentFlag.AlignTop)
        layout.setColumnStretch(1, 1)
        self.setLayout(layout)
        self._key = key
        self.choices = {}
        self.value = MD_LangAlt()
        # label
        if label:
            # put label and lang selector on line above edit box
            layout.addWidget(QtWidgets.QLabel(label), 0, 0)
            edit_pos = 1, 0, 1, 3
            layout.setRowStretch(1, 1)
        else:
            # put lang selector to right of edit box
            edit_pos = 0, 0, 1, 2
        # text edit
        if multi_line:
            self.edit = MultiLineEdit(key, **kw)
        else:
            policy = self.sizePolicy()
            policy.setVerticalPolicy(QtWidgets.QSizePolicy.Policy.Fixed)
            self.setSizePolicy(policy)
            self.edit = SingleLineEdit(key, **kw)
        self.edit.new_value.connect(self._new_value)
        layout.addWidget(self.edit, *edit_pos)
        # language drop down
        self.lang = DropDownSelector(
            'lang', with_multiple=False, extendable=True)
        self.lang.setFixedWidth(width_for_text(self.lang, 'x' * 16))
        self.lang.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.lang.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.lang.new_value.connect(self._change_lang)
        self.lang.customContextMenuRequested.connect(self._context_menu)
        layout.addWidget(self.lang, 0, 2)
        layout.setAlignment(self.lang, Qt.AlignmentFlag.AlignTop)
        # adopt some child methods ...
        self.is_multiple = self.edit.is_multiple
        self.set_height = self.edit.set_height
        # ... and vice versa
        self.lang.define_new_value = self._define_new_lang

    @catch_all
    def setToolTip(self, text):
        self.edit.setToolTip(text)

    @QtSlot(dict)
    @catch_all
    def _change_lang(self, value):
        lang = value['lang']
        if lang == MD_LangAlt.DEFAULT:
            direction = self.layoutDirection()
        else:
            direction = QtCore.QLocale(lang).textDirection()
        if direction == Qt.LayoutDirection.RightToLeft:
            self.edit.set_text_alignment(Qt.AlignmentFlag.AlignRight)
        else:
            self.edit.set_text_alignment(Qt.AlignmentFlag.AlignLeft)
        self.edit.set_value(self.value[lang])

    @QtSlot(dict)
    @catch_all
    def _new_value(self, value):
        value = value[self.edit._key]
        if value in self.choices:
            self.value = self.choices[value]
        else:
            default_lang = self.value.default_lang
            new_value = dict(self.value)
            new_value[self.lang.get_value()] = value
            self.value = MD_LangAlt(
                new_value, default_lang=default_lang, strip=False)
        self.emit_value()

    def _regularise_default(self):
        if not self.value[MD_LangAlt.DEFAULT]:
            return True
        prompt = QtCore.QLocale.system().bcp47Name()
        if prompt in self.value:
            prompt = None
        self.lang.set_value(MD_LangAlt.DEFAULT)
        self.edit.set_value(self.value[MD_LangAlt.DEFAULT])
        lang, OK = QtWidgets.QInputDialog.getText(
            self, translate('LangAltWidget', 'New language'),
            wrap_text(self, translate(
                'LangAltWidget', 'What language is the current text in?'
                ' Please enter an RFC3066 language tag.'), 2), text=prompt)
        if OK and lang:
            self._set_default_lang(lang)
        return OK

    def _define_new_lang(self):
        if not self._regularise_default():
            return None, None
        prompt = QtCore.QLocale.system().bcp47Name()
        if prompt in self.value:
            prompt = None
        lang, OK = QtWidgets.QInputDialog.getText(
            self, translate('LangAltWidget', 'New language'),
            wrap_text(self, translate(
                'LangAltWidget', 'What language would you like to add?'
                ' Please enter an RFC3066 language tag.'), 2), text=prompt)
        if not (OK and lang):
            return None, None
        default_lang = self.value.default_lang
        text = self.value[lang]
        new_value = dict(self.value)
        new_value[lang] = text
        self.value = MD_LangAlt(
            new_value, default_lang=default_lang, strip=False)
        self.emit_value()
        return self.labeled_lang(lang)

    @QtSlot(QtCore.QPoint)
    @catch_all
    def _context_menu(self, pos):
        langs = []
        for n in range(self.lang.count()):
            lang = self.lang.itemData(n)
            if lang and lang != MD_LangAlt.DEFAULT:
                langs.append(lang)
        if not langs:
            return
        default_lang = self.value.default_lang
        menu = QtWidgets.QMenu()
        menu.addAction(translate('LangAltWidget', 'Set default language'))
        for lang in langs:
            action = menu.addAction(lang)
            action.setCheckable(True)
            action.setChecked(lang == default_lang)
        action = execute(menu, self.lang.mapToGlobal(pos))
        if not (action and action.isCheckable()):
            return
        if not self._regularise_default():
            return
        self._set_default_lang(action.text())

    def _set_default_lang(self, lang):
        self.value = MD_LangAlt(self.value, default_lang=lang, strip=False)
        self.emit_value()

    def labeled_lang(self, lang):
        if lang == MD_LangAlt.DEFAULT:
            if len(self.value) == 1:
                return translate('LangAltWidget', 'Language'), lang
            label = '-'
        else:
            label = lang
        label = translate('LangAltWidget', 'Lang: ',
                          'Short abbreviation of "Language: "') + label
        return label, lang

    def set_value(self, value):
        self.choices = {}
        self.lang.setEnabled(True)
        self.value = MD_LangAlt(value, strip=False)
        # use current language, if available
        lang = self.lang.get_value()
        if lang not in self.value:
            # choose language from locale
            lang = QtCore.QLocale.system().bcp47Name()
            if lang not in self.value:
                base_lang = lang.split('-')[0]
                for lang in self.value:
                    if lang.split('-')[0] == base_lang:
                        break
        if lang in self.value:
            lang = self.value.find_key(lang)
        elif self.value:
            # use the default for this value
            lang = self.value.default_lang
        else:
            self.value = MD_LangAlt(default_lang=lang)
        # set language drop down
        self.lang.set_values(
            [self.labeled_lang(x) for x in self.value.languages()],
            default=lang)
        self._change_lang({'lang': lang})

    def get_value(self):
        return self.value

    def set_multiple(self, choices=[]):
        self.choices = {}
        for choice in choices:
            self.choices[str(choice)] = MD_LangAlt(choice, strip=False)
        self.edit.set_multiple(choices=self.choices.keys())
        self.lang.setEnabled(False)


class AugmentSpinBoxBase(WidgetMixin):
    def __init__(self, *arg, **kw):
        self._is_multiple = False
        self._prefix = ''
        self._suffix = ''
        super(AugmentSpinBoxBase, self).__init__(*arg, **kw)
        if self.isRightToLeft():
            self.setAlignment(
                Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        self.set_value(None)
        self.editingFinished.connect(self.emit_value)

    class ContextAction(QtGui2.QAction):
        def __init__(self, value, *arg, **kw):
            super(AugmentSpinBoxBase.ContextAction, self).__init__(*arg, **kw)
            self.setData(value)
            self.triggered.connect(self.set_value)

        @QtSlot()
        @catch_all
        def set_value(self):
            self.parent().set_value(self.data())
            self.parent().emit_value()

    def context_menu_event(self):
        if self.is_multiple() and self.choices:
            QtCore.QTimer.singleShot(0, self.extend_context_menu)

    @QtSlot()
    @catch_all
    def extend_context_menu(self):
        menu = self.findChild(QtWidgets.QMenu)
        if not menu:
            return
        sep = menu.insertSeparator(menu.actions()[0])
        for suggestion in self.choices:
            menu.insertAction(sep, self.ContextAction(
                suggestion, text=self.textFromValue(suggestion), parent=self))

    def fix_up(self):
        if self.is_multiple():
            return True
        if self.cleanText():
            return False
        # user has deleted the value
        self.set_value(None)
        return True

    def init_stepping(self):
        if self.get_value() is None:
            self.setValue(self.default_value)

    def get_value(self):
        value = self.value()
        if value == self.minimum() and self.specialValueText():
            value = None
        return value

    def set_value(self, value):
        self.set_not_multiple()
        if value is None:
            self.setValue(self.minimum())
            self.setSpecialValueText(' ')
        else:
            self.setSpecialValueText('')
            self.setValue(value)


class AugmentDateTime(AugmentSpinBoxBase):
    def set_not_multiple(self):
        if self._is_multiple:
            self._is_multiple = False
            self.set_value(self.default_value)

    def set_multiple(self, choices=[]):
        self.choices = list(filter(None, choices))
        self._is_multiple = True
        self.setValue(self.minimum())
        self.setSpecialValueText(self.multiple)

    def is_multiple(self):
        return self._is_multiple and self.value() == self.minimum()


class AugmentSpinBox(AugmentSpinBoxBase):
    def set_not_multiple(self):
        if self._is_multiple:
            self._is_multiple = False
            self.set_value(self.default_value)
            if self._prefix:
                self.setPrefix(self._prefix)
            if self._suffix:
                self.setSuffix(self._suffix)
            self.lineEdit().setPlaceholderText('')

    def set_multiple(self, choices=[]):
        self.choices = list(filter(None, choices))
        self._is_multiple = True
        if self._prefix:
            self.setPrefix('')
        if self._suffix:
            self.setSuffix('')
        self.lineEdit().setPlaceholderText(self.multiple)
        self.clear()

    def is_multiple(self):
        return self._is_multiple and bool(self.lineEdit().placeholderText())

    def set_prefix(self, prefix):
        self._prefix = prefix
        self.setPrefix(prefix)

    def set_suffix(self, suffix):
        self._suffix = suffix
        self.setSuffix(suffix)


class LatLongDisplay(AugmentSpinBox, QtWidgets.QAbstractSpinBox):
    def __init__(self, *args, **kwds):
        self._key = ('exif:GPSLatitude', 'exif:GPSLongitude')
        self.default_value = ''
        self.multiple = multiple_values()
        super(LatLongDisplay, self).__init__(*args, **kwds)
        self.lat_validator = QtGui.QDoubleValidator(
            -90.0, 90.0, 20, parent=self)
        self.lng_validator = QtGui.QDoubleValidator(
            -180.0, 180.0, 20, parent=self)
        self.setButtonSymbols(self.ButtonSymbols.NoButtons)
        self.label = QtWidgets.QLabel(translate('LatLongDisplay', 'Lat, long'))
        self.label.setAlignment(Qt.AlignmentFlag.AlignRight)
        self.setFixedWidth(width_for_text(self, '8' * 22))
        self.setToolTip('<p>{}</p>'.format(translate(
            'LatLongDisplay', 'Latitude and longitude (in degrees) as two'
            ' decimal numbers separated by a space.')))

    @catch_all
    def focusOutEvent(self, event):
        self.emit_value()
        super(LatLongDisplay, self).focusOutEvent(event)

    @catch_all
    def keyPressEvent(self, event):
        self.set_not_multiple()
        super(LatLongDisplay, self).keyPressEvent(event)

    @catch_all
    def contextMenuEvent(self, event):
        if self.isReadOnly():
            return
        menu = self.lineEdit().createStandardContextMenu()
        suggestion_group = QtGui2.QActionGroup(menu)
        if self._is_multiple and self.choices:
            sep = menu.insertSeparator(menu.actions()[0])
            for suggestion in self.choices:
                label = str(suggestion)
                action = QtGui2.QAction(label, suggestion_group)
                action.setData(str(suggestion))
                menu.insertAction(sep, action)
        action = execute(menu, event.globalPos())
        if action and action.actionGroup() == suggestion_group:
            if self._is_multiple:
                self.set_value(action.data())
                self.emit_value()

    def stepEnabled(self):
        return self.StepEnabledFlag.StepNone

    @catch_all
    def validate(self, text, pos):
        if not text:
            return QtGui.QValidator.State.Acceptable, text, pos
        parts = text.split()
        if len(parts) > 2:
            return QtGui.QValidator.State.Invalid, text, pos
        result = self.lat_validator.validate(parts[0], pos)[0]
        if result != QtGui.QValidator.State.Invalid and len(parts) > 1:
            lng_result = self.lng_validator.validate(parts[1], pos)[0]
            if lng_result == QtGui.QValidator.State.Invalid:
                result = lng_result
            elif result == QtGui.QValidator.State.Acceptable:
                result = lng_result
        return result, text, pos

    @catch_all
    def fixup(self, value):
        value = self.text_to_value(value)
        if len(value) != 2:
            return
        value[0] = min(max(value[0], -90.0), 90.0)
        value[1] = ((value[1] + 180.0) % 360.0) - 180.0
        self.lineEdit().setText(self.value_to_text(value))

    def text_to_value(self, text):
        value = [self.locale().toDouble(x) for x in text.split()]
        if not all(x[1] for x in value):
            # float conversion failed
            return []
        return [x[0] for x in value]

    def value_to_text(self, value):
        return ' '.join(self.locale().toString(float(x), 'f', 6) for x in value)

    def get_value(self):
        value = self.text_to_value(self.text())
        if len(value) != 2:
            return None
        return value

    def get_value_dict(self):
        if self.is_multiple():
            return {}
        return dict(zip(self._key, self.get_value() or (None, None)))

    def set_value(self, value):
        self.set_not_multiple()
        if not value:
            self.clear()
        else:
            if not isinstance(value, str):
                value = self.value_to_text(value)
            self.lineEdit().setText(value)

    def set_value_list(self, values):
        choices = set()
        if not values:
            self.setEnabled(False)
            self.set_value(None)
            return
        self.setEnabled(True)
        for value in values:
            if value:
                lat = value['exif:GPSLatitude']
                lon = value['exif:GPSLongitude']
                if lat and lon:
                    choices.add(self.value_to_text((lat, lon)))
                    continue
            choices.add(None)
        if len(choices) > 1:
            self.set_multiple(choices=[x for x in choices if x])
        else:
            self.set_value(choices and choices.pop())


class DoubleSpinBox(AugmentSpinBox, QtWidgets.QDoubleSpinBox):
    def __init__(self, key, *arg, **kw):
        self._key = key
        self.default_value = 0
        self.multiple = multiple_values()
        super(DoubleSpinBox, self).__init__(*arg, **kw)
        self.setSingleStep(0.1)
        self.setDecimals(4)
        lim = (2 ** 31) - 1
        self.setRange(-lim, lim)
        self.setButtonSymbols(self.ButtonSymbols.NoButtons)

    @catch_all
    def contextMenuEvent(self, event):
        self.context_menu_event()
        return super(DoubleSpinBox, self).contextMenuEvent(event)

    @catch_all
    def keyPressEvent(self, event):
        self.set_not_multiple()
        return super(DoubleSpinBox, self).keyPressEvent(event)

    @catch_all
    def stepBy(self, steps):
        self.set_not_multiple()
        self.init_stepping()
        return super(DoubleSpinBox, self).stepBy(steps)

    @catch_all
    def fixup(self, text):
        if not self.fix_up():
            super(DoubleSpinBox, self).fixup(text)

    @catch_all
    def textFromValue(self, value):
        # don't use QDoubleSpinBox's fixed number of decimals
        decimals = self.decimals()
        value = round(float(value), decimals)
        while decimals > 1 and round(value, decimals - 1) == value:
            decimals -= 1
        return self.locale().toString(value, 'f', decimals)



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

import importlib
import logging
from optparse import OptionParser
import os
import socket
import sys
import warnings

import pkg_resources

from photini import __version__, build
from photini.configstore import BaseConfigStore
from photini.editsettings import EditSettings
from photini.ffmpeg import ffmpeg_version
from photini.imagelist import ImageList
from photini.loggerwindow import LoggerWindow
from photini.metadata import exiv2_version, ImageMetadata
from photini.pyqt import *
from photini.pyqt import QtNetwork, qt_version, qt_version_info
from photini.spelling import SpellCheck, spelling_version

try:
    from photini.gpximporter import GpxImporter
except ImportError as ex:
    print(str(ex))
    GpxImporter = None


logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate


class QTabBar(QtWidgets.QTabBar):
    @catch_all
    def tabSizeHint(self, index):
        size = super(QTabBar, self).tabSizeHint(index)
        if '\n' in self.tabText(index):
            size.setHeight(size.height() + self.fontMetrics().lineSpacing())
        return size


class QTabWidget(QtWidgets.QTabWidget):
    @catch_all
    def resizeEvent(self, event):
        self.adjust_size()
        super(QTabWidget, self).resizeEvent(event)

    def adjust_size(self):
        for idx in range(self.count()):
            self.setTabText(idx, self.tabText(idx).replace('\n', ' '))
        if self.tabBar().sizeHint().width() <= self.width():
            return
        for idx in range(self.count()):
            self.setTabText(idx, wrap_text(self, self.tabText(idx), 2))


class ConfigStore(BaseConfigStore, QtCore.QObject):
    # add timer to save config after it's changed
    def __init__(self, name, *arg, **kw):
        super(ConfigStore, self).__init__(name, *arg, **kw)
        QtCore.QCoreApplication.instance().aboutToQuit.connect(self.save)
        self.timer = QtCore.QTimer(self)
        self.timer.setSingleShot(True)
        self.timer.setInterval(3000)
        self.timer.timeout.connect(self.save)

    def set(self, section, option, value):
        super(ConfigStore, self).set(section, option, value)
        self.timer.start()

    def remove_section(self, section):
        super(ConfigStore, self).remove_section(section)
        self.timer.start()

    @QtSlot()
    @catch_all
    def save(self):
        super(ConfigStore, self).save()


class ServerSocket(QtCore.QObject):
    new_files = QtSignal(list)

    def __init__(self, socket, *arg, **kw):
        super(ServerSocket, self).__init__(*arg, **kw)
        self.socket = socket
        self.data = b''
        self.socket.setParent(self)
        self.socket.readyRead.connect(self.read_data)
        self.socket.disconnected.connect(self.deleteLater)

    @QtSlot()
    @catch_all
    def read_data(self):
        file_list = []
        while self.socket.bytesAvailable():
            self.data += self.socket.readAll().data()
            while b'\n' in self.data:
                line, sep, self.data = self.data.partition(b'\n')
                string = line.decode('utf-8')
                file_list.append(string)
        if file_list:
            self.new_files.emit(file_list)


class InstanceServer(QtNetwork.QTcpServer):
    new_files = QtSignal(list)

    def __init__(self, *arg, **kw):
        super(InstanceServer, self).__init__(*arg, **kw)
        config = BaseConfigStore('instance')
        self.newConnection.connect(self.new_connection)
        if not self.listen(QtNetwork.QHostAddress.SpecialAddress.LocalHost):
            logger.error('Failed to start instance server:', self.errorString())
            return
        config.set('server', 'port', self.serverPort())
        config.save()

    @QtSlot()
    @catch_all
    def new_connection(self):
        window = self.parent().window()
        window.setWindowState(
            (window.windowState() & ~Qt.WindowState.WindowMinimized) |
            Qt.WindowState.WindowActive)
        window.raise_()
        while self.hasPendingConnections():
            socket = self.nextPendingConnection()
            socket = ServerSocket(socket, parent=self)
            socket.new_files.connect(self.new_files)


def SendToInstance(files):
    config = BaseConfigStore('instance')
    port = config.get('server', 'port')
    if not port:
        return False
    try:
        sock = socket.create_connection(('127.0.0.1', int(port)), timeout=1000)
    except ConnectionRefusedError:
        return False
    for path in files:
        data = os.path.abspath(path).encode('utf-8') + b'\n'
        sock.sendall(data)
    sock.close()
    return True


class MenuBar(QtWidgets.QMenuBar):
    def __init__(self, *args, **kwds):
        super(MenuBar, self).__init__(*args, **kwds)
        self.app = QtWidgets.QApplication.instance()
        # file menu
        file_menu = self.addMenu(translate('MenuBar', 'File'))
        action = file_menu.addAction(translate('MenuBar', 'Open files'))
        action.setShortcuts(QtGui.QKeySequence.StandardKey.Open)
        action.triggered.connect(self.app.image_list.open_files)
        self.save_action = file_menu.addAction(
            translate('MenuBar', 'Save changes'))
        self.save_action.setShortcuts(QtGui.QKeySequence.StandardKey.Save)
        self.save_action.setEnabled(False)
        self.save_action.triggered.connect(self.app.image_list.save_files)
        self.fix_thumbs_action = file_menu.addAction(
            translate('MenuBar', 'Fix missing thumbnails'))
        self.fix_thumbs_action.setEnabled(False)
        self.fix_thumbs_action.triggered.connect(
            self.app.image_list.fix_missing_thumbs)
        action = file_menu.addAction(translate('MenuBar', 'Close all files'))
        action.triggered.connect(self.app.image_list.close_all_files)
        file_menu.addSeparator()
        sep = QtWidgets.QWidgetAction(self)
        sep.setDefaultWidget(QtWidgets.QLabel(
            translate('MenuBar', 'Selected images')))
        file_menu.addAction(sep)
        self.selected_actions = self.app.image_list.add_selected_actions(file_menu)
        file_menu.addSeparator()
        action = file_menu.addAction(translate('MenuBar', 'Quit'))
        action.setShortcuts([QtGui.QKeySequence.StandardKey.Quit,
                             QtGui.QKeySequence.StandardKey.Close])
        action.triggered.connect(self.app.closeAllWindows)
        # options menu
        options_menu = self.addMenu(translate('MenuBar', 'Options'))
        action = options_menu.addAction(translate('MenuBar', 'Settings'))
        action.triggered.connect(self.edit_settings)
        options_menu.addSeparator()
        for module in self.parent().modules:
            tab = self.parent().tab_info[module]
            tab['action'] = options_menu.addAction(tab['name'])
            tab['action'].setCheckable(True)
            if tab['class']:
                tab['action'].setChecked(
                    self.app.config_store.get('tabs', module, True))
            else:
                tab['action'].setEnabled(False)
            tab['action'].triggered.connect(self.parent().add_tabs)
        # spelling menu
        languages = self.app.spell_check.available_languages()
        spelling_menu = self.addMenu(translate('MenuBar', 'Spelling'))
        action = spelling_menu.addAction(
            translate('MenuBar', 'Enable spell check'))
        action.setEnabled(languages is not None)
        action.setCheckable(True)
        action.setChecked(self.app.spell_check.enabled)
        action.toggled.connect(self.app.spell_check.enable)
        current_language = self.app.spell_check.current_language()
        if languages:
            language_group = QtGui2.QActionGroup(self)
            for language in sorted(languages):
                language_menu = spelling_menu.addMenu(language)
                for country, code in languages[language]:
                    if country:
                        name = '{}: {}'.format(language, country)
                    else:
                        name = language
                    action = language_menu.addAction(name)
                    action.setCheckable(True)
                    action.setChecked(code == current_language)
                    action.setData(code)
                    action.setActionGroup(language_group)
            language_group.triggered.connect(self.set_language)
        # help menu
        help_menu = self.addMenu(translate('MenuBar', 'Help'))
        action = help_menu.addAction(translate('MenuBar', 'About Photini'))
        action.triggered.connect(self.about)
        action = help_menu.addAction(translate('MenuBar', 'Check for update'))
        action.triggered.connect(self.check_update)
        help_menu.addSeparator()
        action = help_menu.addAction(
            translate('MenuBar', 'Photini documentation'))
        action.triggered.connect(self.open_docs)
        # connect signals
        self.app.image_list.selection_changed.connect(self.new_selection)
        self.app.image_list.image_list_changed.connect(self.new_image_list)
        self.app.image_list.new_metadata.connect(self.new_metadata)

    @QtSlot()
    @catch_all
    def edit_settings(self):
        dialog = EditSettings(self)
        execute(dialog)
        self.parent().tabs.currentWidget().refresh()

    @QtSlot(QtGui2.QAction)
    @catch_all
    def set_language(self, action):
        self.app.spell_check.set_language(action.data())

    @QtSlot()
    @catch_all
    def about(self):
        text = """
<table width="100%"><tr>
<td align="center" width="70%">
<h1>Photini</h1>
<h3>version {0}</h3>
<h4>build {1}</h4>
</td>
<td align="center"><img src="{2}" /></td>
</tr></table>
<p>&copy; Jim Easterbrook <a href="mailto:jim@jim-easterbrook.me.uk">
jim@jim-easterbrook.me.uk</a><br /><br />
{3}<br />
{4}</p>
""".format(__version__, build,
           pkg_resources.resource_filename(
               'photini', 'data/icons/photini_128.png'),
           translate('MenuBar', 'An easy to use digital photograph metadata'
                     ' (Exif, IPTC, XMP) editing application.'),
           translate(
               'MenuBar', 'Open source package available from {url}.').format(
                   url='<a href="https://github.com/jim-easterbrook/Photini">'
                   'github.com/jim-easterbrook/Photini</a>'),
           )
        dialog = QtWidgets.QMessageBox(self)
        dialog.setWindowTitle(translate('MenuBar', 'Photini: about'))
        dialog.setText(text)
        licence = pkg_resources.resource_string('photini', 'data/LICENSE.txt')
        dialog.setDetailedText(licence.decode('utf-8'))
        dialog.setInformativeText(translate(
            'MenuBar', 'This program is released with a GNU General Public'
            ' License. For details click the "{details}" button.').format(
                details=dialog.buttons()[0].text()))
        execute(dialog)

    @QtSlot()
    @catch_all
    def check_update(self):
        import requests
        with Busy():
            try:
                rsp = requests.get(
                    'https://pypi.org/pypi/photini/json', timeout=20)
            except:
                logger.error(str(ex))
                return
        if rsp.status_code != 200:
            logger.error('HTTP error %d', rsp.status_code)
            return
        release = rsp.json()['info']['version']
        dialog = QtWidgets.QMessageBox(self)
        dialog.setWindowTitle(translate('MenuBar', 'Photini: version check'))
        dialog.setText(translate(
            'MenuBar', 'You are currently running Photini version {version}.'
            ' The latest release is {release}.').format(
                version=__version__, release=release))
        execute(dialog)

    @QtSlot()
    @catch_all
    def open_docs(self):
        QtGui.QDesktopServices.openUrl(
            QtCore.QUrl('http://photini.readthedocs.io/'))

    @QtSlot(list)
    @catch_all
    def new_selection(self, selection):
        self.app.image_list.configure_selected_actions(self.selected_actions)

    @QtSlot()
    @catch_all
    def new_image_list(self):
        for image in self.app.image_list.images:
            thumb = image.metadata.thumbnail
            if not thumb or not thumb['image']:
                self.fix_thumbs_action.setEnabled(True)
                return
        self.fix_thumbs_action.setEnabled(False)

    @QtSlot(bool)
    @catch_all
    def new_metadata(self, unsaved_data):
        self.app.image_list.configure_selected_actions(self.selected_actions)
        self.save_action.setEnabled(unsaved_data)


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, options, initial_files):
        super(MainWindow, self).__init__()
        self.setWindowTitle(translate(
            'MenuBar', "Photini photo metadata editor"))
        pixmap = QtGui.QPixmap()
        pixmap.loadFromData(pkg_resources.resource_string(
            'photini', 'data/icons/photini_48.png'))
        icon = QtGui.QIcon(pixmap)
        self.setWindowIcon(icon)
        self.selection = list()
        # logger window
        self.loggerwindow = LoggerWindow(options.verbose)
        self.loggerwindow.setWindowIcon(icon)
        # create shared global objects
        self.app = QtWidgets.QApplication.instance()
        self.app.config_store = ConfigStore('editor', parent=self)
        self.app.spell_check = SpellCheck(parent=self)
        if GpxImporter:
            self.app.gpx_importer = GpxImporter(parent=self)
        else:
            self.app.gpx_importer = None
        self.app.options = options
        self.app.image_list = ImageList()
        self.app.image_list.selection_changed.connect(self.new_selection)
        # initialise metadata handler
        ImageMetadata.initialise(self.app.config_store, options.verbose)
        # restore size and state
        size = self.width(), self.height()
        self.resize(*self.app.config_store.get('main_window', 'size', size))
        window_state = Qt.WindowState(
            self.app.config_store.get('main_window', 'state', 0))
        full_screen = window_state & (
            Qt.WindowState.WindowMaximized | Qt.WindowState.WindowFullScreen)
        if full_screen:
            self.setWindowState(self.windowState() | full_screen)
        # start instance server
        instance_server = InstanceServer(parent=self)
        instance_server.new_files.connect(self.app.image_list.open_file_list,
                                          Qt.ConnectionType.QueuedConnection)
        # prepare list of tabs and associated stuff
        self.tab_info = {}
        default_modules = ['photini.descriptive',  'photini.ownership',
                           'photini.technical',    'photini.regions',
                           'photini.googlemap',    'photini.bingmap',
                           'photini.mapboxmap',    'photini.address',
                           'photini.flickr',       'photini.ipernity',
                           'photini.googlephotos', 'photini.pixelfed',
                           'photini.importer']
        self.modules = self.app.config_store.get(
            'tabs', 'modules', default_modules)
        if 'photini.openstreetmap' in self.modules:
            self.modules.remove('photini.openstreetmap')
            self.app.config_store.delete('tabs', 'photini.openstreetmap')
        self.modules += [x for x in default_modules if x not in self.modules]
        self.app.config_store.set('tabs', 'modules', self.modules)
        for module in self.modules:
            tab = {}
            try:
                mod = importlib.import_module(module)
                tab['class'] = mod.TabWidget
                tab['label'] = tab['class'].tab_name()
                tab['name'] = tab['label'].replace('&', '')
            except ImportError as ex:
                print(str(ex))
                tab['class'] = None
                tab['name'] = module
            self.tab_info[module] = tab
        # menu bar
        self.setMenuBar(MenuBar(parent=self))
        # main application area
        self.central_widget = QtWidgets.QSplitter()
        self.central_widget.setOrientation(Qt.Orientation.Vertical)
        self.central_widget.setChildrenCollapsible(False)
        self.tabs = QTabWidget()
        self.tabs.setTabBar(QTabBar())
        self.tabs.currentChanged.connect(self.new_tab)
        self.tabs.setMovable(True)
        self.tabs.tabBar().tabMoved.connect(self.tab_moved)
        self.add_tabs()
        self.central_widget.addWidget(self.tabs)
        self.central_widget.addWidget(self.app.image_list)
        size = self.central_widget.sizes()
        self.central_widget.setSizes(
            self.app.config_store.get('main_window', 'split', size))
        self.central_widget.splitterMoved.connect(self.new_split)
        self.setCentralWidget(self.central_widget)
        # open files given on command line, after GUI is displayed
        self.initial_files = initial_files
        if self.initial_files:
            QtCore.QTimer.singleShot(0, self.open_initial_files)

    @QtSlot()
    @catch_all
    def open_initial_files(self):
        self.app.image_list.open_file_list(self.initial_files)

    @QtSlot()
    @catch_all
    def add_tabs(self):
        was_blocked = self.tabs.blockSignals(True)
        current = self.tabs.currentWidget()
        self.tabs.clear()
        for module in self.modules:
            tab = self.tab_info[module]
            if not tab['class']:
                self.app.config_store.set('tabs', module, True)
                continue
            use_tab = tab['action'].isChecked()
            self.app.config_store.set('tabs', module, use_tab)
            if not use_tab:
                continue
            if 'object' not in tab:
                tab['object'] = tab['class'](self.app.image_list)
            idx = self.tabs.addTab(tab['object'], tab['label'])
            self.tabs.setTabToolTip(idx, tab['name'])
            self.tabs.tabBar().setTabData(idx, module)
        self.tabs.adjust_size()
        self.tabs.blockSignals(was_blocked)
        if current:
            self.tabs.setCurrentWidget(current)
        self.new_tab(-1)

    @catch_all
    def closeEvent(self, event):
        for n in range(self.tabs.count()):
            if self.tabs.widget(n).do_not_close():
                event.ignore()
                return
        self.app.image_list.unsaved_files_dialog(
            all_files=True, with_cancel=False)
        super(MainWindow, self).closeEvent(event)

    @QtSlot(int, int)
    @catch_all
    def new_split(self, pos, index):
        self.app.config_store.set(
            'main_window', 'split', self.central_widget.sizes())

    @QtSlot(int)
    @catch_all
    def new_tab(self, index):
        current = self.tabs.currentWidget()
        if current:
            self.app.image_list.set_drag_to_map(None)
            current.refresh()

    @QtSlot(int, int)
    @catch_all
    def tab_moved(self, new_pos, old_pos):
        tab_bar = self.tabs.tabBar()
        modules = []
        for n in range(tab_bar.count()):
            modules.append(tab_bar.tabData(n))
        modules += [x for x in self.modules if x not in modules]
        self.modules = modules
        self.app.config_store.set('tabs', 'modules', self.modules)

    @QtSlot(list)
    @catch_all
    def new_selection(self, selection):
        self.tabs.currentWidget().new_selection(selection)

    @catch_all
    def resizeEvent(self, event):
        window_state = self.windowState()
        self.app.config_store.set(
            'main_window', 'state', flag_to_int(window_state))
        if window_state & (Qt.WindowState.WindowMinimized |
                           Qt.WindowState.WindowMaximized |
                           Qt.WindowState.WindowFullScreen):
            return
        size = self.width(), self.height()
        self.app.config_store.set('main_window', 'size', size)


app = None

def main(argv=None):
    global app
    if argv:
        sys.argv = argv
    if qt_version_info < (6, 0):
        QtWidgets.QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
        QtWidgets.QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    # let Qt handle its options
    app = QtWidgets.QApplication(sys.argv)
    # get remaining argument list after Qt has swallowed its options
    sys.argv = app.arguments()
    # install translations
    # English translation as a fallback (to get correct plurals)
    lang_dir = pkg_resources.resource_filename('photini', 'data/lang')
    translator = QtCore.QTranslator(parent=app)
    if translator.load('photini.en', lang_dir):
        app.installTranslator(translator)
        translator = QtCore.QTranslator(parent=app)
    # localised translation, if it exists
    locale = QtCore.QLocale.system()
    if translator.load(locale, 'photini', '.', lang_dir):
        app.installTranslator(translator)
        translator = QtCore.QTranslator(parent=app)
    # parse remaining arguments
    version = 'Photini ' + __version__ + ', build ' + build
    version += '\n  Python ' + sys.version
    version += '\n  ' + exiv2_version
    version += '\n  ' + qt_version
    version += ', locale ' + locale.bcp47Name()
    if spelling_version:
        version += '\n  ' + spelling_version
    if ffmpeg_version:
        version += '\n  ' + ffmpeg_version
    version += '\n  available styles: {}'.format(
        ', '.join(QtWidgets.QStyleFactory.keys()))
    version += '\n  using style: {}'.format(
        QtWidgets.QApplication.style().objectName())
    parser = OptionParser(
        usage=translate('CLIHelp', 'Usage: %prog [options] [file_name, ...]'),
        version=version,
        description=translate('CLIHelp', 'Photini photo metadata editor'))
    parser.add_option(
        '-t', '--test', action='store_true',
        help=translate('CLIHelp', 'test new features or API versions'))
    parser.add_option(
        '-v', '--verbose', action='count', default=0,
        help=translate('CLIHelp', 'increase number of logging messages'))
    options, args = parser.parse_args()
    # if an instance of Photini is already running, send it the list of
    # files to open
    if SendToInstance(args):
        return 0
    # ensure warnings are visible in test mode
    if options.test:
        warnings.simplefilter('default')
    # create GUI and run application event loop
    main = MainWindow(options, args)
    main.show()
    return execute(app)

if __name__ == "__main__":
    sys.exit(main())



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from datetime import datetime
import io
import logging
import os

try:
    import PIL.Image as PIL
except ImportError:
    PIL = None

from photini.ffmpeg import FFmpeg
from photini.metadata import Metadata
from photini.pyqt import *
from photini.pyqt import (image_types, image_types_lower, qt_version_info,
                          set_symbol_font, video_types, video_types_lower)

logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate
DRAG_MIMETYPE = 'application/x-photini-image'


class Image(QtWidgets.QFrame):
    def __init__(self, path, thumb_size=4, *arg, **kw):
        super(Image, self).__init__(*arg, **kw)
        self.app = QtWidgets.QApplication.instance()
        self.path = path
        self.name, ext = os.path.splitext(os.path.basename(self.path))
        self.selected = False
        # read metadata
        self.metadata = Metadata(self.path, notify=self.show_status)
        self.file_times = (os.path.getatime(self.path),
                           os.path.getmtime(self.path))
        # set file type
        self.file_type = self.metadata.mime_type
        # sub widgets
        layout = QtWidgets.QGridLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(3, 3, 3, 3)
        self.setLayout(layout)
        self.setToolTip('<p>' + self.path + '</p>')
        # label to display image
        self.image = QtWidgets.QLabel()
        self.image.setWordWrap(True)
        self.image.setAlignment(
            Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter)
        layout.addWidget(self.image, 0, 0, 1, 2)
        # label to display file name
        self.label = QtWidgets.QLabel()
        self.label.setAlignment(Qt.AlignmentFlag.AlignRight)
        scale_font(self.label, 80)
        layout.addWidget(self.label, 1, 1)
        # label to display status
        self.status = QtWidgets.QLabel()
        self.status.setAlignment(Qt.AlignmentFlag.AlignLeft)
        set_symbol_font(self.status)
        scale_font(self.status, 80)
        layout.addWidget(self.status, 1, 0)
        self.setFrameStyle(
            QtWidgets.QFrame.Shape.Panel | QtWidgets.QFrame.Shadow.Plain)
        self.setObjectName("thumbnail")
        self.set_selected(False)
        self.show_status(False)
        self._set_thumb_size(thumb_size)

    def reload_metadata(self):
        self.metadata = Metadata(self.path, notify=self.show_status)
        self.show_status(False)
        self.load_thumbnail()
        self.app.image_list.emit_selection()

    def transform(self, pixmap, orientation):
        transform = orientation and orientation.get_transform()
        if not transform:
            return pixmap
        return pixmap.transformed(transform)

    def regenerate_thumbnail(self):
        # DCF spec says thumbnail must be 160 x 120, so other aspect
        # ratios are padded with black
        # try using PIL first, good quality and quick
        qt_im = self.get_qt_image()
        if qt_im and PIL:
            data = self.make_thumb_PIL(qt_im)
            if data:
                self.metadata.thumbnail = {'data': data}
                return True
        # next try using FFmpeg, good quality but slower
        data = self.make_thumb_ffmpeg()
        if data:
            self.metadata.thumbnail = {'data': data}
            return True
        # lastly use Qt, quick but not high quality
        if qt_im:
            qt_im = self.make_thumb_Qt(qt_im)
            if qt_im:
                self.metadata.thumbnail = {'image': qt_im}
                return True
        return False

    def make_thumb_ffmpeg(self):
        # get input dimensions
        dims = self.metadata.dimensions
        if not dims:
            return None
        width = dims['width']
        height = dims['height']
        duration = self.metadata.video_duration or 0
        skip = int(min(duration / 2, 10.0))
        # target dimensions
        w, h = 160, 120
        if width < height:
            w, h = h, w
        # use ffmpeg to make scaled, padded, single frame JPEG
        quality = 1
        while True:
            try:
                data = FFmpeg.make_thumbnail(self.path, w, h, skip, quality)
            except Exception as ex:
                logger.error(str(ex))
                return None
            if not data or len(data) < 50000:
                break
            quality += 1
        return data

    def get_qt_image(self):
        reader = QtGui.QImageReader(self.path)
        reader.setAutoTransform(False)
        qt_im = reader.read()
        if not qt_im or qt_im.isNull():
            logger.error('Image read: %s: %s', self.path, reader.errorString())
            return None
        w = qt_im.width()
        h = qt_im.height()
        if max(w, h) > 1000:
            # use Qt's scaling (not high quality) to pre-shrink large
            # images
            qt_im = qt_im.scaled(
                1000, 1000, Qt.AspectRatioMode.KeepAspectRatio,
                Qt.TransformationMode.SmoothTransformation)
            w = qt_im.width()
            h = qt_im.height()
        # pad image to 4:3 (or 3:4) aspect ratio
        if w >= h:
            new_h = int(0.5 + (float(w * 3) / 4.0))
            new_w = int(0.5 + (float(h * 4) / 3.0))
        else:
            new_h = int(0.5 + (float(w * 4) / 3.0))
            new_w = int(0.5 + (float(h * 3) / 4.0))
        if new_w > w:
            pad = (new_w - w) // 2
            qt_im = qt_im.copy(-pad, 0, new_w, h)
        elif new_h > h:
            pad = (new_h - h) // 2
            qt_im = qt_im.copy(0, -pad, w, new_h)
        return qt_im

    def make_thumb_PIL(self, qt_im):
        w, h = 160, 120
        if qt_im.width() < qt_im.height():
            w, h = h, w
        # convert Qt image to PIL image
        buf = QtCore.QBuffer()
        buf.open(buf.OpenModeFlag.WriteOnly)
        qt_im.save(buf, 'PPM')
        data = io.BytesIO(buf.data().data())
        try:
            pil_im = PIL.open(data)
        except Exception as ex:
            logger.error(ex)
            return None
        # scale PIL image
        pil_im.thumbnail((w, h), PIL.ANTIALIAS)
        # save image to memory
        data = io.BytesIO()
        pil_im.save(data, 'JPEG')
        return data.getvalue()

    def make_thumb_Qt(self, qt_im):
        w, h = 160, 120
        if qt_im.width() < qt_im.height():
            w, h = h, w
        # scale Qt image - not as good quality as PIL
        return qt_im.scaled(w, h, Qt.AspectRatioMode.IgnoreAspectRatio,
                            Qt.TransformationMode.SmoothTransformation)

    @catch_all
    def contextMenuEvent(self, event):
        menu = QtWidgets.QMenu(self)
        self.app.image_list.add_selected_actions(menu)
        execute(menu, event.globalPos())

    @catch_all
    def mousePressEvent(self, event):
        if event.button() == Qt.MouseButton.LeftButton:
            if qt_version_info >= (6, 0):
                self.drag_start_pos = event.position()
            else:
                self.drag_start_pos = event.pos()

    @catch_all
    def mouseReleaseEvent(self, event):
        if event.modifiers() == Qt.KeyboardModifier.ControlModifier:
            self.app.image_list.select_image(self, multiple_selection=True)
        elif event.modifiers() == Qt.KeyboardModifier.ShiftModifier:
            self.app.image_list.select_image(self, extend_selection=True)
        else:
            self.app.image_list.select_image(self)

    @catch_all
    def mouseMoveEvent(self, event):
        if not self.app.image_list.drag_icon:
            return
        if qt_version_info >= (6, 0):
            pos = event.position()
        else:
            pos = event.pos()
        if ((pos - self.drag_start_pos).manhattanLength() <
                                    QtWidgets.QApplication.startDragDistance()):
            return
        if not self.get_selected():
            # user has started dragging an unselected image
            self.app.image_list.select_image(self, emit_selection=False)
        paths = []
        for image in self.app.image_list.get_selected_images():
            paths.append(image.path)
        if not paths:
            return
        drag = QtGui.QDrag(self)
        # construct icon
        count = min(len(paths), 8)
        src_icon = self.app.image_list.drag_icon
        src_w = src_icon.width()
        src_h = src_icon.height()
        margin = (count - 1) * 4
        if count == 1:
            icon = src_icon
        else:
            icon = QtGui.QPixmap(src_w + margin, src_h + margin)
            icon.fill(Qt.GlobalColor.transparent)
            try:
                paint = QtGui.QPainter(icon)
                for i in range(count):
                    paint.drawPixmap(
                        QtCore.QPoint(margin - (i * 4), i * 4), src_icon)
            finally:
                del paint
        drag.setPixmap(icon)
        if self.app.image_list.drag_hotspot:
            x, y = self.app.image_list.drag_hotspot
        else:
            x, y = src_w // 2, src_h
        drag.setHotSpot(QtCore.QPoint(x, y + margin))
        mimeData = QtCore.QMimeData()
        mimeData.setData(DRAG_MIMETYPE, repr(paths).encode('utf-8'))
        drag.setMimeData(mimeData)
        if execute(drag,
                   Qt.DropAction.CopyAction) == Qt.DropAction.IgnoreAction:
            # image wasn't dragged to map
            self.app.image_list.emit_selection()

    @catch_all
    def mouseDoubleClickEvent(self, event):
        if event.modifiers() == Qt.KeyboardModifier.NoModifier:
            QtGui.QDesktopServices.openUrl(QtCore.QUrl.fromLocalFile(self.path))

    def show_status(self, changed):
        status = ''
        # set 'geotagged' status
        if self.metadata.gps_info['exif:GPSLatitude']:
            status += chr(0x2690)
        # set 'unsaved' status
        if changed:
            status += chr(0x26A1)
        self.status.setText(status)
        self._elide_name()
        if changed:
            self.app.image_list.new_metadata.emit(True)

    def _elide_name(self):
        self.status.adjustSize()
        elided_name = self.label.fontMetrics().elidedText(
            self.name, Qt.TextElideMode.ElideLeft,
            self.image.width() - self.status.width())
        self.label.setText(elided_name)

    def _set_thumb_size(self, thumb_size):
        width = width_for_text(self.label, 'X' * thumb_size * 20) // 6
        self.image.setFixedSize(width, width)
        self._elide_name()

    def set_thumb_size(self, thumb_size):
        self._set_thumb_size(thumb_size)
        self.load_thumbnail()

    def load_thumbnail(self, label=None):
        label = label or self.image
        image = self.metadata.thumbnail and self.metadata.thumbnail['image']
        if not image:
            label.setText(wrap_text(
                label, translate('ImageList', 'No thumbnail in file'), lines=4))
            return
        pixmap = QtGui.QPixmap.fromImage(image)
        pixmap = self.transform(pixmap, self.metadata.orientation)
        rect = label.contentsRect()
        label.setPixmap(
            pixmap.scaled(rect.width(), rect.height(),
                          Qt.AspectRatioMode.KeepAspectRatio,
                          Qt.TransformationMode.SmoothTransformation))

    def set_selected(self, value):
        self.selected = value
        if self.selected:
            self.setStyleSheet("#thumbnail {border: 2px solid red}")
        else:
            self.setStyleSheet("#thumbnail {border: 2px solid grey}")

    def get_selected(self):
        return self.selected


class ScrollArea(QtWidgets.QScrollArea):
    dropped_images = QtSignal(list)

    def __init__(self, image_list=None, parent=None):
        super(ScrollArea, self).__init__(parent)
        self.image_list = image_list
        self.setWidgetResizable(True)
        self.setAcceptDrops(True)
        widget = QtWidgets.QWidget()
        self.thumbs = ThumbsLayout(scroll_area=self)
        widget.setLayout(self.thumbs)
        self.setWidget(widget)
        # adopt some layout methods & signals
        self.add_widget = self.thumbs.addWidget
        self.remove_widget = self.thumbs.removeWidget

    @catch_all
    def ensureWidgetVisible(self, widget):
        left, top, right, bottom = self.thumbs.getContentsMargins()
        super(ScrollArea, self).ensureWidgetVisible(
            widget, max(left, right), max(top, bottom))

    @catch_all
    def dropEvent(self, event):
        file_list = []
        for uri in event.mimeData().urls():
            file_list.append(uri.toLocalFile())
        if file_list:
            self.dropped_images.emit(file_list)

    @catch_all
    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat('text/uri-list'):
            event.acceptProposedAction()

    @catch_all
    def resizeEvent(self, event):
        super(ScrollArea, self).resizeEvent(event)
        self.thumbs.do_layout()

    def usable_size(self):
        rect = self.contentsRect()
        width = rect.width() - self.verticalScrollBar().sizeHint().width()
        height = rect.height() - self.horizontalScrollBar().sizeHint().height()
        return width, height

    def set_minimum_height(self, min_height):
        bar = self.horizontalScrollBar()
        if bar.isVisible():
            min_height += bar.height()
        margins = self.contentsMargins()
        self.setMinimumHeight(min_height + margins.top() + margins.bottom())

    @QtSlot()
    @catch_all
    def multi_row_changed(self):
        if self.image_list.last_selected:
            self.ensureWidgetVisible(self.image_list.last_selected)


class ThumbsLayout(QtWidgets.QLayout):
    """Multi-row fixed-width or single-row variable-width grid of
    thumbnail widgets, according to height.

    """
    def __init__(self, scroll_area=None, **kw):
        super(ThumbsLayout, self).__init__(**kw)
        self.scroll_area = scroll_area
        self.item_list = []
        self.multi_row = None
        self.do_layout()

    def addItem(self, item):
        self.item_list.append(item)

    def horizontalSpacing(self):
        return 0

    def verticalSpacing(self):
        return 0

    def count(self):
        return len(self.item_list)

    def itemAt(self, idx):
        if idx < 0 or idx >= len(self.item_list):
            return None
        return self.item_list[idx]

    def takeAt(self, idx):
        if idx < 0 or idx >= len(self.item_list):
            return None
        return self.item_list.pop(idx)

    def expandingDirections(self):
        return 0

    def hasHeightForWidth(self):
        return False

    @catch_all
    def setGeometry(self, rect):
        super(ThumbsLayout, self).setGeometry(rect)
        self.do_layout(rect)

    def sizeHint(self):
        return self.size_hint

    def minimumSize(self):
        return self.size_hint

    def do_layout(self, rect=None):
        left, top, right, bottom = self.getContentsMargins()
        width_hint = left + right
        height_hint = top + bottom
        if self.item_list and self.scroll_area:
            item_size = self.item_list[0].sizeHint()
            item_h = item_size.height()
            item_w = item_size.width()
            row_height = item_h + height_hint
            self.scroll_area.set_minimum_height(row_height)
            view_width, view_height = self.scroll_area.usable_size()
            multi_row = view_height > row_height
            if multi_row != self.multi_row:
                self.multi_row = multi_row
                # make selected item visible after redrawing has finished
                QtCore.QTimer.singleShot(0, self.scroll_area.multi_row_changed)
            if multi_row:
                columns = max((view_width - width_hint) // item_w, 1)
                rows = (len(self.item_list) + columns - 1) // columns
            else:
                columns = len(self.item_list)
                rows = 1
            width_hint += columns * item_w
            height_hint += rows * item_h
        self.size_hint = QtCore.QSize(width_hint, height_hint)
        widget = self.parentWidget()
        if widget:
            widget.setMinimumSize(self.size_hint)
        if not (rect and self.item_list):
            return
        if QtWidgets.QApplication.isRightToLeft():
            x = rect.right() - right - item_w
            item_w = -item_w
        else:
            x = rect.left() + left
        y = rect.top() + top
        for n, item in enumerate(self.item_list):
            i, j = n % columns, n // columns
            item.setGeometry(QtCore.QRect(
                QtCore.QPoint(x + (i * item_w), y + (j * item_h)), item_size))


class ImageList(QtWidgets.QWidget):
    image_list_changed = QtSignal()
    new_metadata = QtSignal(bool)
    selection_changed = QtSignal(list)
    sort_order_changed = QtSignal()

    def __init__(self, parent=None):
        super(ImageList, self).__init__(parent)
        self.app = QtWidgets.QApplication.instance()
        self.drag_icon = None
        self.images = []
        self.last_selected = None
        self.selection_anchor = None
        self.thumb_size = self.app.config_store.get('controls', 'thumb_size', 4)
        if self.thumb_size > 20:
            # old config, in pixels
            self.thumb_size = self.thumb_size // 20
        # thumbnail display
        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.scroll_area = ScrollArea(image_list=self)
        self.scroll_area.dropped_images.connect(self.open_file_list)
        self.layout().addWidget(self.scroll_area)
        QtGui2.QShortcut(QtGui.QKeySequence.StandardKey.MoveToPreviousChar,
                         self.scroll_area, self.move_to_prev_thumb)
        QtGui2.QShortcut(QtGui.QKeySequence.StandardKey.MoveToNextChar,
                         self.scroll_area, self.move_to_next_thumb)
        QtGui2.QShortcut(QtGui.QKeySequence.StandardKey.MoveToStartOfLine,
                         self.scroll_area, self.move_to_first_thumb)
        QtGui2.QShortcut(QtGui.QKeySequence.StandardKey.MoveToEndOfLine,
                         self.scroll_area, self.move_to_last_thumb)
        QtGui2.QShortcut(QtGui.QKeySequence.StandardKey.SelectPreviousChar,
                         self.scroll_area, self.select_prev_thumb)
        QtGui2.QShortcut(QtGui.QKeySequence.StandardKey.SelectNextChar,
                         self.scroll_area, self.select_next_thumb)
        QtGui2.QShortcut(QtGui.QKeySequence.StandardKey.SelectAll,
                         self.scroll_area, self.select_all)
        # sort key selector
        bottom_bar = QtWidgets.QHBoxLayout()
        self.layout().addLayout(bottom_bar)
        bottom_bar.addWidget(QtWidgets.QLabel(
            translate('ImageList', 'Sort by')))
        self.sort_name = QtWidgets.QRadioButton(
            translate('ImageList', 'file name'))
        self.sort_name.clicked.connect(self._new_sort_order)
        bottom_bar.addWidget(self.sort_name)
        self.sort_date = QtWidgets.QRadioButton(
            translate('ImageList', 'date taken'))
        self.sort_date.clicked.connect(self._new_sort_order)
        bottom_bar.addWidget(self.sort_date)
        if self.app.config_store.get('controls', 'sort_date', False):
            self.sort_date.setChecked(True)
        else:
            self.sort_name.setChecked(True)
        # size selector
        bottom_bar.addStretch(1)
        bottom_bar.addWidget(QtWidgets.QLabel(
            translate('ImageList', 'Thumbnail size')))
        self.size_slider = QtWidgets.QSlider(Qt.Orientation.Horizontal)
        self.size_slider.setTracking(False)
        self.size_slider.setRange(4, 9)
        self.size_slider.setPageStep(1)
        self.size_slider.setValue(self.thumb_size)
        self.size_slider.setTickPosition(
            QtWidgets.QSlider.TickPosition.TicksBelow)
        self.size_slider.setMinimumWidth(
            width_for_text(self.size_slider, 'x' * 20))
        self.size_slider.valueChanged.connect(self._new_thumb_size)
        bottom_bar.addWidget(self.size_slider)

    def set_drag_to_map(self, icon, hotspot=None):
        self.drag_icon = icon
        self.drag_hotspot = hotspot

    def get_image(self, path):
        for image in self.images:
            if image.path == path:
                return image
        return None

    def get_images(self):
        return self.images

    @catch_all
    def mousePressEvent(self, event):
        if self.scroll_area.underMouse():
            self._clear_selection()
            self.last_selected = None
            self.selection_anchor = None
            self.emit_selection()

    @QtSlot(bool)
    @catch_all
    def open_files(self, checked=False):
        args = [
            self,
            translate('ImageList', 'Open files'),
            self.app.config_store.get('paths', 'images', ''),
            translate('ImageList',
                      "Images ({0});;Videos ({1});;All files (*)").format(
                          ' '.join(['*.' + x for x in image_types()]),
                          ' '.join(['*.' + x for x in video_types()]))
            ]
        if not self.app.config_store.get('pyqt', 'native_dialog', True):
            args += [None, QtWidgets.QFileDialog.Option.DontUseNativeDialog]
        path_list = QtWidgets.QFileDialog.getOpenFileNames(*args)
        path_list = path_list[0]
        if not path_list:
            return
        self.open_file_list(path_list)

    @QtSlot(list)
    @catch_all
    def open_file_list(self, path_list, top_level=True, dir_list=[]):
        last_path = None
        types = ['.' + x for x in (image_types_lower() + video_types_lower())]
        with Busy():
            for path in path_list:
                if os.path.basename(path).startswith('.'):
                    # don't open .directory or .thumbs
                    continue
                if os.path.isdir(path):
                    path = os.path.realpath(path)
                    if path in dir_list:
                        # don't open directories we've already opened
                        continue
                    dir_list.append(path)
                    last_path = self.open_file_list(
                        [os.path.join(path, x) for x in os.listdir(path)],
                        top_level=False, dir_list=dir_list) or last_path
                elif (not top_level and
                          os.path.splitext(path)[1].lower() not in types):
                    pass
                elif self.open_file(path):
                    last_path = path
        if top_level and last_path:
            self.done_opening(last_path)
        return last_path

    def open_file(self, path):
        path = os.path.realpath(path)
        base, ext = os.path.splitext(path)
        if ext.lower() == '.xmp':
            if os.path.isfile(base):
                path = base
            else:
                dir_name = os.path.dirname(path)
                for file in os.listdir(dir_name):
                    path = os.path.join(dir_name, file)
                    b, e = os.path.splitext(path)
                    if b == base and e.lower() != '.xmp':
                        break
                else:
                    return False
        if not os.path.isfile(path):
            return False
        if self.get_image(path):
            # already opened this path
            return True
        image = Image(path, thumb_size=self.thumb_size)
        self.images.append(image)
        self.show_thumbnail(image)
        return True

    def done_opening(self, path):
        self.app.config_store.set(
            'paths', 'images', os.path.dirname(os.path.abspath(path)))
        self._sort_thumbnails()

    def _date_key(self, image):
        result = (image.metadata.date_taken or image.metadata.date_digitised
                  or image.metadata.date_modified)
        if result:
            result = result['datetime']
        else:
            # use file date as last resort
            result = datetime.fromtimestamp(os.path.getmtime(image.path))
        # convert result to string and append path so photos with same
        # time stamp get sorted consistently
        result = result.strftime('%Y%m%d%H%M%S%f') + image.path
        return result

    @QtSlot()
    @catch_all
    def _new_sort_order(self):
        self._sort_thumbnails()
        self.sort_order_changed.emit()

    def _sort_thumbnails(self):
        sort_date = self.sort_date.isChecked()
        self.app.config_store.set('controls', 'sort_date', sort_date)
        with Busy():
            if sort_date:
                self.images.sort(key=self._date_key)
            else:
                self.images.sort(key=lambda x: x.path)
            for image in self.images:
                self.show_thumbnail(image, False)
        if self.last_selected:
            self.app.processEvents()
            self.scroll_area.ensureWidgetVisible(self.last_selected)
        self.image_list_changed.emit()

    def show_thumbnail(self, image, live=True):
        self.scroll_area.add_widget(image)
        if live:
            self.app.processEvents()
        image.load_thumbnail()
        if live:
            self.app.processEvents()
            self.scroll_area.ensureWidgetVisible(image)
            self.app.processEvents()

    def add_selected_actions(self, menu):
        actions = {}
        actions['reload'] = menu.addAction('', self.reload_selected_metadata)
        actions['save'] = menu.addAction(
            translate('ImageList', 'Save changes'), self.save_selected_metadata)
        actions['diff'] = menu.addAction(
            translate('ImageList', 'View changes'), self.diff_selected_metadata)
        actions['thumbs'] = menu.addAction(
            '', self.regenerate_selected_thumbnails)
        actions['close'] = menu.addAction('', self.close_selected_files)
        self.configure_selected_actions(actions)
        return actions

    def configure_selected_actions(self, actions):
        images = self.get_selected_images()
        changed_images = any([x.metadata.changed() for x in images])
        actions['reload'].setEnabled(bool(images))
        actions['save'].setEnabled(changed_images)
        actions['diff'].setEnabled(changed_images)
        actions['thumbs'].setEnabled(bool(images))
        actions['close'].setEnabled(bool(images))

        if qt_version_info >= (6, 0):
            # pyside6-lupdate doesn't recognise plurals with 'translate'
            actions['reload'].setText(
                ImageList.tr('Reload file(s)', '', len(images)))
            actions['thumbs'].setText(
                ImageList.tr('Regenerate thumbnail(s)', '', len(images)))
            actions['close'].setText(
                ImageList.tr('Close file(s)', '', len(images)))
        else:
            # Qt5 doesn't handle ClassName.tr correctly
            actions['reload'].setText(translate(
                'ImageList', 'Reload file(s)', '', len(images)))
            actions['thumbs'].setText(translate(
                'ImageList', 'Regenerate thumbnail(s)', '', len(images)))
            actions['close'].setText(translate(
                'ImageList', 'Close file(s)', '', len(images)))

    @QtSlot()
    @catch_all
    def reload_selected_metadata(self):
        with Busy():
            for image in self.get_selected_images():
                image.reload_metadata()

    @QtSlot()
    @catch_all
    def save_selected_metadata(self):
        self._save_files(images=self.get_selected_images())

    @QtSlot()
    @catch_all
    def diff_selected_metadata(self):
        dialog = QtWidgets.QDialog(parent=self)
        dialog.setLayout(QtWidgets.QVBoxLayout())
        width = width_for_text(dialog, 'x' * 120)
        dialog.setFixedSize(min(width, self.window().width()),
                            min(width // 2, self.window().height()))
        table = QtWidgets.QTableWidget()
        table.setVerticalScrollMode(table.ScrollMode.ScrollPerPixel)
        table.setColumnCount(3)
        table.setHorizontalHeaderLabels([translate('ImageList', 'new value'),
                                         translate('ImageList', 'undo'),
                                         translate('ImageList', 'old value')])
        table.horizontalHeader().setSectionResizeMode(
            0, QtWidgets.QHeaderView.ResizeMode.Stretch)
        table.horizontalHeader().setSectionResizeMode(
            2, QtWidgets.QHeaderView.ResizeMode.Stretch)
        dialog.layout().addWidget(table)
        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.StandardButton.Ok |
            QtWidgets.QDialogButtonBox.StandardButton.Cancel)
        button_box.accepted.connect(dialog.accept)
        button_box.rejected.connect(dialog.reject)
        dialog.layout().addWidget(button_box)
        changed = False
        position = None
        for image in self.get_selected_images():
            if not image.metadata.changed():
                continue
            dialog.setWindowTitle(translate(
                'ImageList', 'Metadata differences: {file_name}').format(
                    file_name=image.name))
            labels = []
            row = 0
            undo = {}
            table.clearContents()
            new_md = image.metadata
            old_md = Metadata(image.path)
            for key in ('title', 'headline', 'description', 'alt_text',
                        'alt_text_ext', 'keywords', 'rating',
                        'creator', 'creator_title', 'credit_line', 'copyright',
                        'rights', 'instructions', 'contact_info',
                        'date_taken', 'date_digitised', 'date_modified',
                        'orientation', 'camera_model', 'lens_model',
                        'focal_length', 'focal_length_35', 'aperture',
                        'gps_info', 'location_taken', 'location_shown',
                        'image_region', 'thumbnail'):
                values = getattr(new_md, key), getattr(old_md, key)
                if values[0] == values[1]:
                    continue
                values = [str(x or '') for x in values]
                table.setRowCount(row + 1)
                for n, value in enumerate(values):
                    item = QtWidgets.QTableWidgetItem(value)
                    table.setItem(row, n * 2, item)
                undo[key] = QtWidgets.QTableWidgetItem()
                undo[key].setFlags(
                    undo[key].flags() | Qt.ItemFlag.ItemIsUserCheckable)
                undo[key].setCheckState(Qt.CheckState.Unchecked)
                table.setItem(row, 1, undo[key])
                labels.append(key)
                row += 1
            if not row:
                continue
            table.setVerticalHeaderLabels(labels)
            table.resizeColumnsToContents()
            table.resizeRowsToContents()
            if position:
                dialog.move(position)
            result = execute(dialog)
            if result != QtWidgets.QDialog.DialogCode.Accepted:
                return
            position = dialog.pos()
            undo_all = True
            for key, widget in undo.items():
                if widget.checkState() == Qt.CheckState.Checked:
                    setattr(new_md, key, getattr(old_md, key))
                    changed = True
                else:
                    undo_all = False
            if undo_all:
                image.reload_metadata()
        if changed:
            self.emit_selection()

    @QtSlot()
    @catch_all
    def regenerate_selected_thumbnails(self):
        with Busy():
            for image in self.get_selected_images():
                if image.regenerate_thumbnail():
                    image.load_thumbnail()
                    self.app.processEvents()

    @QtSlot()
    @catch_all
    def fix_missing_thumbs(self):
        with Busy():
            for image in self.get_images():
                thumb = image.metadata.thumbnail
                if not thumb or not thumb['image']:
                    if image.regenerate_thumbnail():
                        image.load_thumbnail()
                        self.app.processEvents()
        self.image_list_changed.emit()

    @QtSlot()
    @catch_all
    def close_selected_files(self):
        self.close_files(False)

    @QtSlot()
    @catch_all
    def close_all_files(self):
        self.close_files(True)

    def close_files(self, all_files):
        if not self.unsaved_files_dialog(all_files=all_files):
            return
        if all_files:
            close_list = list(self.images)
        else:
            close_list = self.get_selected_images()
        if not close_list:
            return
        idx = self.images.index(close_list[0])
        for image in close_list:
            self.images.remove(image)
            self.scroll_area.remove_widget(image)
            image.setParent(None)
        if 0 <= idx < len(self.images):
            self.select_image(self.images[idx])
        else:
            self.last_selected = None
            self.selection_anchor = None
            self.emit_selection()
        self.image_list_changed.emit()

    @QtSlot(bool)
    @catch_all
    def save_files(self, checked=False):
        self._save_files(self.images)

    def _save_files(self, images=[]):
        self._flush_editing()
        if_mode = self.app.config_store.get('files', 'image', True)
        sc_mode = self.app.config_store.get('files', 'sidecar', 'auto')
        iptc_mode = self.app.config_store.get('files', 'iptc_iim', 'preserve')
        keep_time = self.app.config_store.get(
            'files', 'preserve_timestamps', 'now')
        if isinstance(keep_time, bool):
            # old config format
            keep_time = ('now', 'keep')[keep_time]
        if not images:
            images = self.images
        with Busy():
            for image in images:
                if keep_time == 'taken' and image.metadata.date_taken:
                    file_times = (
                        image.file_times[0],
                        image.metadata.date_taken['datetime'].timestamp())
                elif keep_time == 'keep':
                    file_times = image.file_times
                else:
                    file_times = None
                image.metadata.save(
                    if_mode=if_mode, sc_mode=sc_mode,
                    iptc_mode=iptc_mode, file_times=file_times)
        unsaved = any([image.metadata.changed() for image in self.images])
        self.new_metadata.emit(unsaved)

    def unsaved_files_dialog(
            self, all_files=False, with_cancel=True, with_discard=True):
        """Return true if OK to continue with close or quit or whatever"""
        self._flush_editing()
        for image in self.images:
            if image.metadata.changed() and (all_files or image.selected):
                break
        else:
            return True
        dialog = QtWidgets.QMessageBox(parent=self)
        dialog.setWindowTitle(translate('ImageList', 'Photini: unsaved data'))
        dialog.setText('<h3>{}</h3>'.format(
            translate('ImageList', 'Some images have unsaved metadata.')))
        dialog.setInformativeText(
            translate('ImageList', 'Do you want to save your changes?'))
        dialog.setIcon(dialog.Icon.Warning)
        buttons = dialog.StandardButton.Save
        if with_cancel:
            buttons |= dialog.StandardButton.Cancel
        if with_discard:
            buttons |= dialog.StandardButton.Discard
        dialog.setStandardButtons(buttons)
        dialog.setDefaultButton(dialog.StandardButton.Save)
        result = execute(dialog)
        if result == dialog.StandardButton.Save:
            self._save_files()
            return True
        return result == dialog.StandardButton.Discard

    def _flush_editing(self):
        # finish any editing in progress
        current_focus = self.app.focusWidget()
        if current_focus:
            current_focus.clearFocus()

    def get_selected_images(self):
        selection = []
        for image in self.images:
            if image.get_selected():
                selection.append(image)
        return selection

    def emit_selection(self):
        self.selection_changed.emit(self.get_selected_images())

    def select_all(self):
        for image in self.images:
            image.set_selected(True)
        self.selection_anchor = None
        self.last_selected = None
        self.emit_selection()

    def move_to_prev_thumb(self):
        self._inc_selection(-1)

    def move_to_next_thumb(self):
        self._inc_selection(1)

    def move_to_first_thumb(self):
        self.select_image(self.images[0])

    def move_to_last_thumb(self):
        self.select_image(self.images[-1])

    def select_prev_thumb(self):
        self._inc_selection(-1, extend_selection=True)

    def select_next_thumb(self):
        self._inc_selection(1, extend_selection=True)

    def _inc_selection(self, inc, extend_selection=False):
        if self.last_selected:
            idx = self.images.index(self.last_selected)
            idx = (idx + inc) % len(self.images)
        else:
            idx = 0
        self.select_image(self.images[idx], extend_selection=extend_selection)

    @QtSlot(int)
    @catch_all
    def _new_thumb_size(self, value):
        self.thumb_size = value
        self.app.config_store.set('controls', 'thumb_size', self.thumb_size)
        for image in self.images:
            image.set_thumb_size(self.thumb_size)
        if self.last_selected:
            self.app.processEvents()
            self.scroll_area.ensureWidgetVisible(self.last_selected)

    def select_image(self, image, extend_selection=False,
                     multiple_selection=False, emit_selection=True):
        self.scroll_area.ensureWidgetVisible(image)
        if extend_selection and self.selection_anchor:
            idx1 = self.images.index(self.selection_anchor)
            idx2 = self.images.index(self.last_selected)
            for i in range(min(idx1, idx2), max(idx1, idx2) + 1):
                self.images[i].set_selected(False)
            idx2 = self.images.index(image)
            for i in range(min(idx1, idx2), max(idx1, idx2) + 1):
                self.images[i].set_selected(True)
        elif multiple_selection:
            image.set_selected(not image.get_selected())
            self.selection_anchor = image
        else:
            self._clear_selection()
            image.set_selected(True)
            self.selection_anchor = image
        self.last_selected = image
        if emit_selection:
            self.emit_selection()

    def select_images(self, images):
        self._clear_selection()
        if not images:
            self.last_selected = None
            self.selection_anchor = None
            self.emit_selection()
            return
        for image in images:
            image.set_selected(True)
            self.scroll_area.ensureWidgetVisible(image)
        self.selection_anchor = images[0]
        self.last_selected = images[-1]
        self.emit_selection()

    def _clear_selection(self):
        for image in self.images:
            if image.get_selected():
                image.set_selected(False)



# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-21  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import ast
import codecs
from configparser import RawConfigParser
import os
import pprint
import stat
import sys

import appdirs
import pkg_resources

class BaseConfigStore(object):
    # the actual config store functionality
    def __init__(self, name, *arg, **kw):
        super(BaseConfigStore, self).__init__(*arg, **kw)
        self.dirty = False
        self.config = RawConfigParser()
        config_dir = os.environ.get('PHOTINI_CONFIG')
        if config_dir:
            config_dir = os.path.expanduser(config_dir)
        elif hasattr(appdirs, 'user_config_dir'):
            config_dir = appdirs.user_config_dir('photini')
        else:
            config_dir = appdirs.user_data_dir('photini')
        if not os.path.isdir(config_dir):
            os.makedirs(config_dir, mode=stat.S_IRWXU)
        self.file_name = os.path.join(config_dir, name + '.ini')
        if os.path.isfile(self.file_name):
            kwds = {'encoding': 'utf-8'}
            with open(self.file_name, 'r', **kwds) as fp:
                self.config.read_file(fp)
        self.has_section = self.config.has_section

    def get(self, section, option, default=None):
        if self.config.has_option(section, option):
            value = self.config.get(section, option)
            if not value:
                return None
            try:
                value = ast.literal_eval(value)
            except Exception:
                pass
            return value
        if default is not None:
            self.set(section, option, default)
        return default

    def set(self, section, option, value):
        value = pprint.pformat(value)
        if not self.config.has_section(section):
            self.config.add_section(section)
        elif (self.config.has_option(section, option) and
              self.config.get(section, option) == value):
            return
        self.config.set(section, option, value)
        self.dirty = True

    def delete(self, section, option):
        if not self.config.has_section(section):
            return
        if self.config.has_option(section, option):
            self.config.remove_option(section, option)
        if not self.config.options(section):
            self.config.remove_section(section)
        self.dirty = True

    def remove_section(self, section):
        if not self.config.has_section(section):
            return
        for option in self.config.options(section):
            self.config.remove_option(section, option)
        self.config.remove_section(section)
        self.dirty = True

    def save(self):
        if not self.dirty:
            return
        kwds = {'encoding': 'utf-8'}
        with open(self.file_name, 'w', **kwds) as fp:
            self.config.write(fp)
        os.chmod(self.file_name, stat.S_IRUSR | stat.S_IWUSR)
        self.dirty = False


class KeyStore(object):
    """Store OAuth2 client ids and client 'secrets'.

    Google recognise that client secrets can't be kept secret in an
    application that runs on a user's computer. See
    https://developers.google.com/identity/protocols/OAuth2InstalledApp
    for more background. However, they also say the secret "may not be
    embedded in open source projects" (see section 4.b.1 of
    https://developers.google.com/terms/).

    Photini stores the client credentials in a separate file, using mild
    obfuscation to hide the actual values. If this is insufficient to
    satisfy Google then the keys file will have to be removed from open
    source and distributed by other means. Or users will need to create
    their own by registering as a developer at Google.

    The position with Flickr keys is less clear, but there's no harm in
    obfuscating them as well.

    """
    def __init__(self):
        self.config = RawConfigParser()
        if sys.version_info >= (3, 2):
            data = pkg_resources.resource_string('photini', 'data/keys.txt')
            data = data.decode('utf-8')
            self.config.read_string(data)
        else:
            data = pkg_resources.resource_stream('photini', 'data/keys.txt')
            self.config.readfp(data)

    def get(self, section, option):
        value = self.config.get(section, option)
        value = value.encode('ascii')
        value = codecs.decode(value, 'base64_codec')
        return value.decode('ascii')


# create single object for entire application
key_store = KeyStore()



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

import locale
import logging

import requests

from photini.configstore import key_store
from photini.photinimap import GeocoderBase, PhotiniMap
from photini.pyqt import Busy, Qt, QtCore, QtWidgets, scale_font
from photini.widgets import Label

logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate


class GoogleGeocoder(GeocoderBase):
    api_key = key_store.get('googlemap', 'api_key')
    interval = 50

    def query(self, params, url):
        params['key'] = self.api_key
        with Busy():
            self.rate_limit()
            try:
                rsp = requests.get(url, params=params, timeout=5)
            except Exception as ex:
                logger.error(str(ex))
                return []
        rsp = rsp.json()
        if rsp['status'] != 'OK':
            if 'error_message' in rsp:
                logger.error(
                    'Search error: %s: %s', rsp['status'], rsp['error_message'])
            else:
                logger.error('Search error: %s', rsp['status'])
            return []
        results = rsp['results']
        if not results:
            logger.error('No results found')
            return []
        return results

    def get_altitude(self, coords):
        params = {'locations': '{:.5f},{:.5f}'.format(*coords)}
        results = self.cached_query(
            params, 'https://maps.googleapis.com/maps/api/elevation/json')
        if results:
            return results[0]['elevation']
        return None

    def search(self, search_string, bounds=None):
        params = {'address': search_string}
        lang, encoding = locale.getdefaultlocale()
        if lang:
            params['language'] = lang
        if bounds:
            north, east, south, west = bounds
            params['bounds'] = '{:.4f},{:.4f}|{:.4f},{:.4f}'.format(
                south, west, north, east)
        for result in self.cached_query(
                params, 'https://maps.googleapis.com/maps/api/geocode/json'):
            bounds = result['geometry']['viewport']
            yield (bounds['northeast']['lat'], bounds['northeast']['lng'],
                   bounds['southwest']['lat'], bounds['southwest']['lng'],
                   result['formatted_address'])

    def search_terms(self):
        widget = Label(translate(
            'MapTabGoogle', 'Search and altitude lookup powered by Google',
            'Do not translate "powered by Google"'), lines=2)
        widget.setAlignment(Qt.AlignmentFlag.AlignRight)
        scale_font(widget, 80)
        return [widget]


class TabWidget(PhotiniMap):
    api_key = key_store.get('googlemap', 'api_key')

    @staticmethod
    def tab_name():
        return translate('MapTabGoogle', 'Map (&Google)')

    def get_geocoder(self):
        return GoogleGeocoder(parent=self)

    def get_head(self):
        url = 'http://maps.googleapis.com/maps/api/js?callback=initialize'
        if self.app.options.test:
            url += '&v=beta'
        url += '&key=' + self.api_key
        lang, encoding = locale.getdefaultlocale()
        if lang:
            language, sep, region = lang.replace('_', '-').partition('-')
            url += '&language=' + language
            if region:
                url += '&region=' + region
        return '''    <script type="text/javascript"
      src="{}" async>
    </script>'''.format(url)



#  Photini - a simple photo metadata editor.
#  http://github.com/jim-easterbrook/Photini
#  Copyright (C) 2012-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
#
#  This program is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see
#  <http://www.gnu.org/licenses/>.

import logging

from photini.pyqt import (
    catch_all, FormLayout, Qt, QtCore, QtSlot, QtWidgets, width_for_text)

logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate


class EditSettings(QtWidgets.QDialog):
    def __init__(self, *arg, **kw):
        super(EditSettings, self).__init__(*arg, **kw)
        self.app = QtWidgets.QApplication.instance()
        self.config_store = self.app.config_store
        self.setWindowTitle(translate('EditSettings', 'Photini: settings'))
        self.setLayout(QtWidgets.QVBoxLayout())
        # main dialog area
        scroll_area = QtWidgets.QScrollArea()
        self.layout().addWidget(scroll_area)
        panel = QtWidgets.QWidget()
        layout = FormLayout()
        panel.setLayout(layout)
        if layout.rowWrapPolicy() == layout.RowWrapPolicy.DontWrapRows:
            layout.setRowWrapPolicy(layout.RowWrapPolicy.WrapLongRows)
        # apply & cancel buttons
        self.button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.StandardButton.Apply |
            QtWidgets.QDialogButtonBox.StandardButton.Cancel)
        self.button_box.clicked.connect(self.button_clicked)
        self.layout().addWidget(self.button_box)
        # IPTC data
        if self.config_store.get('files', 'force_iptc'):
            self.config_store.set('files', 'iptc_iim', 'create')
        self.config_store.delete('files', 'force_iptc')
        iptc_mode = self.config_store.get('files', 'iptc_iim', 'preserve')
        button_group = QtWidgets.QButtonGroup(parent=self)
        self.iptc_always = QtWidgets.QRadioButton(
            translate('EditSettings', 'Always write'))
        button_group.addButton(self.iptc_always)
        self.iptc_always.setChecked(iptc_mode == 'create')
        self.iptc_always.setMinimumWidth(
            width_for_text(self.iptc_always, 'x' * 35))
        panel.layout().addRow(
            translate('EditSettings', 'IPTC-IIM metadata'), self.iptc_always)
        self.iptc_auto = QtWidgets.QRadioButton(
            translate('EditSettings', 'Write if exists'))
        button_group.addButton(self.iptc_auto)
        self.iptc_auto.setChecked(iptc_mode == 'preserve')
        panel.layout().addRow('', self.iptc_auto)
        self.iptc_delete = QtWidgets.QRadioButton(
            translate('EditSettings', 'Always delete'))
        button_group.addButton(self.iptc_delete)
        self.iptc_delete.setChecked(iptc_mode == 'delete')
        panel.layout().addRow('', self.iptc_delete)
        # show IPTC-IIM length limits
        length_warning = self.config_store.get('files', 'length_warning', True)
        self.length_warning = QtWidgets.QCheckBox(
            translate('EditSettings', 'Show IPTC-IIM length limits'))
        self.length_warning.setChecked(length_warning)
        panel.layout().addRow('', self.length_warning)
        # sidecar files
        if_mode = self.config_store.get('files', 'image', True)
        sc_mode = self.config_store.get('files', 'sidecar', 'auto')
        if not if_mode:
            sc_mode = 'always'
        button_group = QtWidgets.QButtonGroup(parent=self)
        self.sc_always = QtWidgets.QRadioButton(
            translate('EditSettings', 'Always create'))
        button_group.addButton(self.sc_always)
        self.sc_always.setChecked(sc_mode == 'always')
        panel.layout().addRow(
            translate('EditSettings', 'Sidecar files'), self.sc_always)
        self.sc_auto = QtWidgets.QRadioButton(
            translate('EditSettings', 'Create if necessary'))
        button_group.addButton(self.sc_auto)
        self.sc_auto.setChecked(sc_mode == 'auto')
        self.sc_auto.setEnabled(if_mode)
        panel.layout().addRow('', self.sc_auto)
        self.sc_delete = QtWidgets.QRadioButton(
            translate('EditSettings', 'Delete when possible'))
        button_group.addButton(self.sc_delete)
        self.sc_delete.setChecked(sc_mode == 'delete')
        self.sc_delete.setEnabled(if_mode)
        panel.layout().addRow('', self.sc_delete)
        # image file locking
        self.write_if = QtWidgets.QCheckBox(
            translate('EditSettings', '(when possible)'))
        self.write_if.setChecked(if_mode)
        self.write_if.clicked.connect(self.new_write_if)
        panel.layout().addRow(
            translate('EditSettings', 'Write to image file'), self.write_if)
        # preserve file timestamps
        keep_time = self.config_store.get('files', 'preserve_timestamps', 'now')
        if isinstance(keep_time, bool):
            # old config format
            keep_time = ('now', 'keep')[keep_time]
        button_group = QtWidgets.QButtonGroup(parent=self)
        self.keep_time = QtWidgets.QRadioButton(
            translate('EditSettings', 'Keep original'))
        button_group.addButton(self.keep_time)
        self.keep_time.setChecked(keep_time=='keep')
        panel.layout().addRow(
            translate('EditSettings', 'File timestamps'), self.keep_time)
        self.time_taken = QtWidgets.QRadioButton(
            translate('EditSettings', 'Set to when photo was taken'))
        button_group.addButton(self.time_taken)
        self.time_taken.setChecked(keep_time=='taken')
        panel.layout().addRow('', self.time_taken)
        button = QtWidgets.QRadioButton(
            translate('EditSettings', 'Set to when file is saved'))
        button_group.addButton(button)
        button.setChecked(keep_time=='now')
        panel.layout().addRow('', button)
        # import GPX altitude
        if self.app.gpx_importer:
            self.gpx_altitude = QtWidgets.QCheckBox(
                translate('EditSettings', 'set altitude'))
            self.gpx_altitude.setChecked(
                self.config_store.get('map', 'gpx_altitude', True))
            panel.layout().addRow(
                translate('EditSettings', 'GPX importer'), self.gpx_altitude)
        # add panel to scroll area after its size is known
        scroll_area.setWidget(panel)

    @QtSlot()
    @catch_all
    def new_write_if(self):
        if_mode = self.write_if.isChecked()
        self.sc_auto.setEnabled(if_mode)
        self.sc_delete.setEnabled(if_mode)
        if not if_mode:
            self.sc_always.setChecked(True)

    @QtSlot(QtWidgets.QAbstractButton)
    @catch_all
    def button_clicked(self, button):
        if button != self.button_box.button(
                QtWidgets.QDialogButtonBox.StandardButton.Apply):
            return self.reject()
        # change config
        if self.iptc_always.isChecked():
            iptc_mode = 'create'
        elif self.iptc_auto.isChecked():
            iptc_mode = 'preserve'
        else:
            iptc_mode = 'delete'
        self.config_store.set('files', 'iptc_iim', iptc_mode)
        self.config_store.set(
            'files', 'length_warning', self.length_warning.isChecked())
        if self.sc_always.isChecked():
            sc_mode = 'always'
        elif self.sc_auto.isChecked():
            sc_mode = 'auto'
        else:
            sc_mode = 'delete'
        self.config_store.set('files', 'sidecar', sc_mode)
        self.config_store.set('files', 'image', self.write_if.isChecked())
        if self.keep_time.isChecked():
            keep_time = 'keep'
        elif self.time_taken.isChecked():
            keep_time = 'taken'
        else:
            keep_time = 'now'
        self.config_store.set('files', 'preserve_timestamps', keep_time)
        if self.app.gpx_importer:
            self.config_store.set(
                'map', 'gpx_altitude', self.gpx_altitude.isChecked())
        return self.accept()
